# Copyright (c) 2015 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Cinder Volume driver for Fujitsu ETERNUS DX S3 series."""

import ast
import base64
import time

from lxml import etree as ET
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils.secretutils import md5
from oslo_utils import units
import six

from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder.volume import configuration as conf
from cinder.volume.drivers.fujitsu.eternus_dx import constants as CONSTANTS
from cinder.volume.drivers.fujitsu.eternus_dx import eternus_dx_cli
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

try:
    import pywbem
    pywbemAvailable = True
except ImportError:
    pywbemAvailable = False

FJ_ETERNUS_DX_OPT_opts = [
    cfg.StrOpt('cinder_eternus_config_file',
               default='/etc/cinder/cinder_fujitsu_eternus_dx.xml',
               help='Config file for cinder eternus_dx volume driver.'),
]

CONF.register_opts(FJ_ETERNUS_DX_OPT_opts, group=conf.SHARED_CONF_GROUP)


class FJDXCommon(object):
    """Common code that does not depend on protocol."""

    VERSION = "1.3.0"
    stats = {
        'driver_version': VERSION,
        'storage_protocol': None,
        'vendor_name': 'FUJITSU',
        'QoS_support': False,
        'volume_backend_name': None,
    }

    def __init__(self, prtcl, configuration=None):

        self.pywbemAvailable = pywbemAvailable

        self.protocol = prtcl
        self.configuration = configuration
        self.configuration.append_config_values(FJ_ETERNUS_DX_OPT_opts)

        if prtcl == 'iSCSI':
            # Get iSCSI ipaddress from driver configuration file.
            self.configuration.iscsi_ip_address = (
                self._get_drvcfg('EternusISCSIIP'))
        self.conn = None
        self.fjdxcli = {}
        self._check_user()

    @staticmethod
    def get_driver_options():
        return FJ_ETERNUS_DX_OPT_opts

    def create_volume(self, volume):
        """Create volume on ETERNUS."""
        LOG.debug('create_volume, '
                  'volume id: %(vid)s, volume size: %(vsize)s.',
                  {'vid': volume['id'], 'vsize': volume['size']})

        self.conn = self._get_eternus_connection()
        volumesize = int(volume['size']) * units.Gi
        volumename = self._create_volume_name(volume['id'])

        LOG.debug('create_volume, volumename: %(volumename)s, '
                  'volumesize: %(volumesize)u.',
                  {'volumename': volumename,
                   'volumesize': volumesize})

        # get poolname from driver configuration file
        eternus_pool = volume_utils.extract_host(volume['host'], 'pool')
        # Existence check the pool
        pool = self._find_pool(eternus_pool)

        if 'RSP' in pool['InstanceID']:
            pooltype = CONSTANTS.RAIDGROUP
        else:
            pooltype = CONSTANTS.TPPOOL

        configservice = self._find_eternus_service(CONSTANTS.STOR_CONF)
        if configservice is None:
            msg = (_('create_volume, volume: %(volume)s, '
                     'volumename: %(volumename)s, '
                     'eternus_pool: %(eternus_pool)s, '
                     'Storage Configuration Service not found.')
                   % {'volume': volume,
                      'volumename': volumename,
                      'eternus_pool': eternus_pool})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('create_volume, '
                  'CreateOrModifyElementFromStoragePool, '
                  'ConfigService: %(service)s, '
                  'ElementName: %(volumename)s, '
                  'InPool: %(eternus_pool)s, '
                  'ElementType: %(pooltype)u, '
                  'Size: %(volumesize)u.',
                  {'service': configservice,
                   'volumename': volumename,
                   'eternus_pool': eternus_pool,
                   'pooltype': pooltype,
                   'volumesize': volumesize})

        # Invoke method for create volume
        rc, errordesc, job = self._exec_eternus_service(
            'CreateOrModifyElementFromStoragePool',
            configservice,
            ElementName=volumename,
            InPool=pool,
            ElementType=self._pywbem_uint(pooltype, '16'),
            Size=self._pywbem_uint(volumesize, '64'))

        if rc == CONSTANTS.VOLUMENAME_IN_USE:  # Element Name is in use
            LOG.warning('create_volume, '
                        'volumename: %(volumename)s, '
                        'Element Name is in use.',
                        {'volumename': volumename})
            vol_instance = self._find_lun(volume)
            element = vol_instance
        elif rc != 0:
            msg = (_('create_volume, '
                     'volumename: %(volumename)s, '
                     'poolname: %(eternus_pool)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'volumename': volumename,
                      'eternus_pool': eternus_pool,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        else:
            element = job['TheElement']

        # Get eternus model name
        try:
            systemnamelist = (
                self._enum_eternus_instances('FUJITSU_StorageProduct'))
        except Exception:
            msg = (_('create_volume, '
                     'volume: %(volume)s, '
                     'EnumerateInstances, '
                     'cannot connect to ETERNUS.')
                   % {'volume': volume})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('create_volume, '
                  'volumename: %(volumename)s, '
                  'Return code: %(rc)lu, '
                  'Error: %(errordesc)s, '
                  'Backend: %(backend)s, '
                  'Pool Name: %(eternus_pool)s, '
                  'Pool Type: %(pooltype)s.',
                  {'volumename': volumename,
                   'rc': rc,
                   'errordesc': errordesc,
                   'backend': systemnamelist[0]['IdentifyingNumber'],
                   'eternus_pool': eternus_pool,
                   'pooltype': CONSTANTS.POOL_TYPE_dic[pooltype]})

        # Create return value.
        element_path = {
            'classname': element.classname,
            'keybindings': {
                'CreationClassName': element['CreationClassName'],
                'SystemName': element['SystemName'],
                'DeviceID': element['DeviceID'],
                'SystemCreationClassName': element['SystemCreationClassName']
            }
        }

        volume_no = "0x" + element['DeviceID'][24:28]

        metadata = {'FJ_Backend': systemnamelist[0]['IdentifyingNumber'],
                    'FJ_Volume_Name': volumename,
                    'FJ_Volume_No': volume_no,
                    'FJ_Pool_Name': eternus_pool,
                    'FJ_Pool_Type': CONSTANTS.POOL_TYPE_dic[pooltype]}

        return (element_path, metadata)

    def create_pool_info(self, pool_instance, volume_count, pool_type):
        """Create pool information from pool instance."""
        LOG.debug('create_pool_info, pool_instance: %(pool)s, '
                  'volume_count: %(volcount)s, pool_type: %(ptype)s.',
                  {'pool': pool_instance,
                   'volcount': volume_count, 'ptype': pool_type})

        if pool_type not in CONSTANTS.POOL_TYPE_list:
            msg = (_('Invalid pool type was specified : %s.') % pool_type)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        total_gb = pool_instance['TotalManagedSpace'] / units.Gi
        free_gb = pool_instance['RemainingManagedSpace'] / units.Gi

        if hasattr(pool_instance, 'provisioned_capacity_gb'):
            prov_gb = pool_instance.provisioned_capacity_gb
        else:
            prov_gb = total_gb - free_gb

        if pool_type == 'RAID':
            useable_gb = free_gb
        else:
            # If the ratio is less than the value on ETERNUS,
            # useable_gb may be negative. Avoid over-allocation.
            max_capacity = total_gb * float(
                self.configuration.max_over_subscription_ratio)
            useable_gb = max_capacity - prov_gb

        pool = {
            'name': pool_instance['ElementName'],
            'path': pool_instance.path,
            'total_capacity_gb': total_gb,
            'free_capacity_gb': free_gb,
            'type': pool_type,
            'volume_count': volume_count,
            'provisioned_capacity_gb': prov_gb,
            'useable_capacity_gb': useable_gb
        }

        LOG.debug('create_pool_info, pool: %s.', pool)
        return pool

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""
        LOG.debug('create_volume_from_snapshot, '
                  'volume id: %(vid)s, volume size: %(vsize)s, '
                  'snapshot id: %(sid)s.',
                  {'vid': volume['id'], 'vsize': volume['size'],
                   'sid': snapshot['id']})

        self.conn = self._get_eternus_connection()
        source_volume_instance = self._find_lun(snapshot)

        # Check the existence of source volume.
        if source_volume_instance is None:
            msg = _('create_volume_from_snapshot, '
                    'Source Volume does not exist in ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Create volume for the target volume.
        (element_path, metadata) = self.create_volume(volume)
        target_volume_instancename = self._create_eternus_instance_name(
            element_path['classname'], element_path['keybindings'])

        try:
            target_volume_instance = (
                self._get_eternus_instance(target_volume_instancename))
        except Exception:
            msg = (_('create_volume_from_snapshot, '
                     'target volume instancename: %(volume_instancename)s, '
                     'Get Instance Failed.')
                   % {'volume_instancename': target_volume_instancename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        self._create_local_cloned_volume(target_volume_instance,
                                         source_volume_instance)

        return (element_path, metadata)

    def create_cloned_volume(self, volume, src_vref):
        """Create clone of the specified volume."""
        LOG.debug('create_cloned_volume, '
                  'tgt: (%(tid)s, %(tsize)s), src: (%(sid)s, %(ssize)s).',
                  {'tid': volume['id'], 'tsize': volume['size'],
                   'sid': src_vref['id'], 'ssize': src_vref['size']})

        self.conn = self._get_eternus_connection()
        source_volume_instance = self._find_lun(src_vref)

        if source_volume_instance is None:
            msg = _('create_cloned_volume, '
                    'Source Volume does not exist in ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        (element_path, metadata) = self.create_volume(volume)
        target_volume_instancename = self._create_eternus_instance_name(
            element_path['classname'], element_path['keybindings'])

        try:
            target_volume_instance = (
                self._get_eternus_instance(target_volume_instancename))
        except Exception:
            msg = (_('create_cloned_volume, '
                     'target volume instancename: %(volume_instancename)s, '
                     'Get Instance Failed.')
                   % {'volume_instancename': target_volume_instancename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        self._create_local_cloned_volume(target_volume_instance,
                                         source_volume_instance)

        return (element_path, metadata)

    @lockutils.synchronized('ETERNUS-vol', 'cinder-', True)
    def _create_local_cloned_volume(self, tgt_vol_instance, src_vol_instance):
        """Create local clone of the specified volume."""
        s_volumename = src_vol_instance['ElementName']
        t_volumename = tgt_vol_instance['ElementName']

        LOG.debug('_create_local_cloned_volume, '
                  'tgt volume name: %(t_volumename)s, '
                  'src volume name: %(s_volumename)s, ',
                  {'t_volumename': t_volumename,
                   's_volumename': s_volumename})

        # Get replication service for CreateElementReplica.
        repservice = self._find_eternus_service(CONSTANTS.REPL)

        if repservice is None:
            msg = _('_create_local_cloned_volume, '
                    'Replication Service not found.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Invoke method for create cloned volume from volume.
        rc, errordesc, job = self._exec_eternus_service(
            'CreateElementReplica',
            repservice,
            SyncType=self._pywbem_uint(8, '16'),
            SourceElement=src_vol_instance.path,
            TargetElement=tgt_vol_instance.path)

        if rc != 0:
            msg = (_('_create_local_cloned_volume, '
                     'volumename: %(volumename)s, '
                     'sourcevolumename: %(sourcevolumename)s, '
                     'source volume instance: %(source_volume)s, '
                     'target volume instance: %(target_volume)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'volumename': t_volumename,
                      'sourcevolumename': s_volumename,
                      'source_volume': src_vol_instance.path,
                      'target_volume': tgt_vol_instance.path,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_create_local_cloned_volume, out: %(rc)s, %(job)s.',
                  {'rc': rc, 'job': job})

    def delete_volume(self, volume):
        """Delete volume on ETERNUS."""
        LOG.debug('delete_volume, volume id: %s.', volume['id'])

        self.conn = self._get_eternus_connection()
        vol_exist = self._delete_volume_setting(volume)

        if not vol_exist:
            LOG.debug('delete_volume, volume not found in 1st check.')
            return False

        # Check volume existence on ETERNUS again
        # because volume is deleted when SnapOPC copysession is deleted.
        vol_instance = self._find_lun(volume)
        if vol_instance is None:
            LOG.debug('delete_volume, volume not found in 2nd check, '
                      'but no problem.')
            return True

        self._delete_volume(vol_instance)
        return True

    @lockutils.synchronized('ETERNUS-vol', 'cinder-', True)
    def _delete_volume_setting(self, volume):
        """Delete volume setting (HostAffinity, CopySession) on ETERNUS."""
        LOG.debug('_delete_volume_setting, volume id: %s.', volume['id'])

        # Check the existence of volume.
        volumename = self._create_volume_name(volume['id'])
        vol_instance = self._find_lun(volume)

        if vol_instance is None:
            LOG.info('_delete_volume_setting, volumename:%(volumename)s, '
                     'volume not found on ETERNUS.',
                     {'volumename': volumename})
            return False

        # Delete host-affinity setting remained by unexpected error.
        self._unmap_lun(volume, None, force=True)

        # Check copy session relating to target volume.
        cpsessionlist = self._find_copysession(vol_instance)
        delete_copysession_list = []
        wait_copysession_list = []

        for cpsession in cpsessionlist:
            LOG.debug('_delete_volume_setting, '
                      'volumename: %(volumename)s, '
                      'cpsession: %(cpsession)s.',
                      {'volumename': volumename,
                       'cpsession': cpsession})

            if cpsession['SyncedElement'] == vol_instance.path:
                # Copy target : other_volume --(copy)--> vol_instance
                delete_copysession_list.append(cpsession)
            elif cpsession['SystemElement'] == vol_instance.path:
                # Copy source : vol_instance --(copy)--> other volume
                wait_copysession_list.append(cpsession)

        LOG.debug('_delete_volume_setting, '
                  'wait_cpsession: %(wait_cpsession)s, '
                  'delete_cpsession: %(delete_cpsession)s.',
                  {'wait_cpsession': wait_copysession_list,
                   'delete_cpsession': delete_copysession_list})

        for cpsession in wait_copysession_list:
            self._wait_for_copy_complete(cpsession)

        for cpsession in delete_copysession_list:
            self._delete_copysession(cpsession)

        LOG.debug('_delete_volume_setting, '
                  'wait_cpsession: %(wait_cpsession)s, '
                  'delete_cpsession: %(delete_cpsession)s, complete.',
                  {'wait_cpsession': wait_copysession_list,
                   'delete_cpsession': delete_copysession_list})
        return True

    @lockutils.synchronized('ETERNUS-vol', 'cinder-', True)
    def _delete_volume(self, vol_instance):
        """Delete volume on ETERNUS."""
        LOG.debug('_delete_volume, volume name: %s.',
                  vol_instance['ElementName'])

        volumename = vol_instance['ElementName']

        configservice = self._find_eternus_service(CONSTANTS.STOR_CONF)
        if configservice is None:
            msg = (_('_delete_volume, volumename: %(volumename)s, '
                     'Storage Configuration Service not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_delete_volume, volumename: %(volumename)s, '
                  'vol_instance: %(vol_instance)s, '
                  'Method: ReturnToStoragePool.',
                  {'volumename': volumename,
                   'vol_instance': vol_instance.path})

        # Invoke method for delete volume
        rc, errordesc, job = self._exec_eternus_service(
            'ReturnToStoragePool',
            configservice,
            TheElement=vol_instance.path)

        if rc != 0:
            msg = (_('_delete_volume, volumename: %(volumename)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_delete_volume, volumename: %(volumename)s, '
                  'Return code: %(rc)lu, '
                  'Error: %(errordesc)s.',
                  {'volumename': volumename,
                   'rc': rc,
                   'errordesc': errordesc})

    @lockutils.synchronized('ETERNUS-vol', 'cinder-', True)
    def create_snapshot(self, snapshot):
        """Create snapshot using SnapOPC."""
        LOG.debug('create_snapshot, '
                  'snapshot id: %(sid)s, volume id: %(vid)s.',
                  {'sid': snapshot['id'], 'vid': snapshot['volume_id']})

        self.conn = self._get_eternus_connection()
        snapshotname = snapshot['name']
        volumename = snapshot['volume_name']
        vol_id = snapshot['volume_id']
        volume = snapshot['volume']
        d_volumename = self._create_volume_name(snapshot['id'])
        s_volumename = self._create_volume_name(vol_id)
        vol_instance = self._find_lun(volume)
        repservice = self._find_eternus_service(CONSTANTS.REPL)

        # Check the existence of volume.
        if vol_instance is None:
            # Volume not found on ETERNUS.
            msg = (_('create_snapshot, '
                     'volumename: %(s_volumename)s, '
                     'source volume not found on ETERNUS.')
                   % {'s_volumename': s_volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if repservice is None:
            msg = (_('create_snapshot, '
                     'volumename: %(volumename)s, '
                     'Replication Service not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Get poolname from driver configuration file.
        eternus_pool = self._get_drvcfg('EternusSnapPool')
        # Check the existence of pool
        pool = self._find_pool(eternus_pool)
        if pool is None:
            msg = (_('create_snapshot, '
                     'eternus_pool: %(eternus_pool)s, '
                     'pool not found.')
                   % {'eternus_pool': eternus_pool})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('create_snapshot, '
                  'snapshotname: %(snapshotname)s, '
                  'source volume name: %(volumename)s, '
                  'vol_instance.path: %(vol_instance)s, '
                  'dest_volumename: %(d_volumename)s, '
                  'pool: %(pool)s, '
                  'Invoke CreateElementReplica.',
                  {'snapshotname': snapshotname,
                   'volumename': volumename,
                   'vol_instance': vol_instance.path,
                   'd_volumename': d_volumename,
                   'pool': pool})

        # Invoke method for create snapshot
        rc, errordesc, job = self._exec_eternus_service(
            'CreateElementReplica',
            repservice,
            ElementName=d_volumename,
            TargetPool=pool,
            SyncType=self._pywbem_uint(7, '16'),
            SourceElement=vol_instance.path)

        if rc != 0:
            msg = (_('create_snapshot, '
                     'snapshotname: %(snapshotname)s, '
                     'source volume name: %(volumename)s, '
                     'vol_instance.path: %(vol_instance)s, '
                     'dest volume name: %(d_volumename)s, '
                     'pool: %(pool)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'snapshotname': snapshotname,
                      'volumename': volumename,
                      'vol_instance': vol_instance.path,
                      'd_volumename': d_volumename,
                      'pool': pool,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        else:
            element = job['TargetElement']

        LOG.debug('create_snapshot, '
                  'volumename:%(volumename)s, '
                  'Return code:%(rc)lu, '
                  'Error:%(errordesc)s.',
                  {'volumename': volumename,
                   'rc': rc,
                   'errordesc': errordesc})

        # Create return value.
        element_path = {
            'classname': element.classname,
            'keybindings': {
                'CreationClassName': element['CreationClassName'],
                'SystemName': element['SystemName'],
                'DeviceID': element['DeviceID'],
                'SystemCreationClassName': element['SystemCreationClassName']
            }
        }

        sdv_no = "0x" + element['DeviceID'][24:28]
        metadata = {'FJ_SDV_Name': d_volumename,
                    'FJ_SDV_No': sdv_no,
                    'FJ_Pool_Name': eternus_pool}
        return (element_path, metadata)

    def delete_snapshot(self, snapshot):
        """Delete snapshot."""
        LOG.debug('delete_snapshot, '
                  'snapshot id: %(sid)s, volume id: %(vid)s.',
                  {'sid': snapshot['id'], 'vid': snapshot['volume_id']})

        vol_exist = self.delete_volume(snapshot)
        LOG.debug('delete_snapshot, vol_exist: %s.', vol_exist)
        return vol_exist

    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info."""
        LOG.debug('initialize_connection, '
                  'volume id: %(vid)s, protocol: %(prtcl)s.',
                  {'vid': volume['id'], 'prtcl': self.protocol})

        self.conn = self._get_eternus_connection()
        vol_instance = self._find_lun(volume)
        # Check the existence of volume
        if vol_instance is None:
            # Volume not found
            msg = (_('initialize_connection, '
                     'volume: %(volume)s, '
                     'Volume not found.')
                   % {'volume': volume['name']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        target_portlist = self._get_target_port()
        mapdata = self._get_mapdata(vol_instance, connector, target_portlist)

        if mapdata:
            # volume is already mapped
            target_lun = mapdata.get('target_lun', None)
            target_luns = mapdata.get('target_luns', None)

            LOG.info('initialize_connection, '
                     'volume: %(volume)s, '
                     'target_lun: %(target_lun)s, '
                     'target_luns: %(target_luns)s, '
                     'Volume is already mapped.',
                     {'volume': volume['name'],
                      'target_lun': target_lun,
                      'target_luns': target_luns})
        else:
            self._map_lun(vol_instance, connector, target_portlist)
            mapdata = self._get_mapdata(vol_instance,
                                        connector, target_portlist)

        mapdata['target_discovered'] = True
        mapdata['volume_id'] = volume['id']

        if self.protocol == 'fc':
            device_info = {'driver_volume_type': 'fibre_channel',
                           'data': mapdata}
        elif self.protocol == 'iSCSI':
            device_info = {'driver_volume_type': 'iscsi',
                           'data': mapdata}

        LOG.debug('initialize_connection, '
                  'device_info:%(info)s.',
                  {'info': device_info})
        return device_info

    def terminate_connection(self, volume, connector, force=False, **kwargs):
        """Disallow connection from connector."""
        LOG.debug('terminate_connection, '
                  'volume id: %(vid)s, protocol: %(prtcl)s, force: %(frc)s.',
                  {'vid': volume['id'], 'prtcl': self.protocol, 'frc': force})

        self.conn = self._get_eternus_connection()
        force = True if not connector else force
        map_exist = self._unmap_lun(volume, connector, force)

        LOG.debug('terminate_connection, map_exist: %s.', map_exist)
        return map_exist

    def build_fc_init_tgt_map(self, connector, target_wwn=None):
        """Build parameter for Zone Manager"""
        LOG.debug('build_fc_init_tgt_map, target_wwn: %s.', target_wwn)

        initiatorlist = self._find_initiator_names(connector)

        if target_wwn is None:
            target_wwn = []
            target_portlist = self._get_target_port()
            for target_port in target_portlist:
                target_wwn.append(target_port['Name'])

        init_tgt_map = {initiator: target_wwn for initiator in initiatorlist}

        LOG.debug('build_fc_init_tgt_map, '
                  'initiator target mapping: %s.', init_tgt_map)
        return init_tgt_map

    def check_attached_volume_in_zone(self, connector):
        """Check Attached Volume in Same FC Zone or not"""
        LOG.debug('check_attached_volume_in_zone, connector: %s.', connector)

        aglist = self._find_affinity_group(connector)
        if not aglist:
            attached = False
        else:
            attached = True

        LOG.debug('check_attached_volume_in_zone, attached: %s.', attached)
        return attached

    @lockutils.synchronized('ETERNUS-vol', 'cinder-', True)
    def extend_volume(self, volume, new_size):
        """Extend volume on ETERNUS."""
        LOG.debug('extend_volume, volume id: %(vid)s, '
                  'size: %(size)s, new_size: %(nsize)s.',
                  {'vid': volume['id'],
                   'size': volume['size'], 'nsize': new_size})

        self.conn = self._get_eternus_connection()
        volumesize = new_size * units.Gi
        volumename = self._create_volume_name(volume['id'])

        # Get source volume instance.
        vol_instance = self._find_lun(volume)
        if vol_instance is None:
            msg = (_('extend_volume, '
                     'volumename: %(volumename)s, '
                     'volume not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('extend_volume, volumename: %(volumename)s, '
                  'volumesize: %(volumesize)u, '
                  'volume instance: %(vol_instance)s.',
                  {'volumename': volumename,
                   'volumesize': volumesize,
                   'vol_instance': vol_instance.path})

        # Get poolname from driver configuration file.
        pool_name, pool = self._find_pool_from_volume(vol_instance)
        if pool is None:
            msg = (_('extend_volume, '
                     'eternus_pool: %(eternus_pool)s, '
                     'pool not found.')
                   % {'eternus_pool': pool_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Set pooltype.
        if 'RSP' in pool['InstanceID']:
            pooltype = CONSTANTS.RAIDGROUP
        else:
            pooltype = CONSTANTS.TPPOOL

        configservice = self._find_eternus_service(CONSTANTS.STOR_CONF)
        if not configservice:
            msg = (_('extend_volume, volume: %(volume)s, '
                     'volumename: %(volumename)s, '
                     'eternus_pool: %(eternus_pool)s, '
                     'Storage Configuration Service not found.')
                   % {'volume': volume,
                      'volumename': volumename,
                      'eternus_pool': pool_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('extend_volume, '
                  'CreateOrModifyElementFromStoragePool, '
                  'ConfigService: %(service)s, '
                  'ElementName: %(volumename)s, '
                  'InPool: %(eternus_pool)s, '
                  'ElementType: %(pooltype)u, '
                  'Size: %(volumesize)u, '
                  'TheElement: %(vol_instance)s.',
                  {'service': configservice,
                   'volumename': volumename,
                   'eternus_pool': pool_name,
                   'pooltype': pooltype,
                   'volumesize': volumesize,
                   'vol_instance': vol_instance.path})

        # Invoke method for extend volume
        rc, errordesc, job = self._exec_eternus_service(
            'CreateOrModifyElementFromStoragePool',
            configservice,
            ElementName=volumename,
            InPool=pool,
            ElementType=self._pywbem_uint(pooltype, '16'),
            Size=self._pywbem_uint(volumesize, '64'),
            TheElement=vol_instance.path)

        if rc != 0:
            msg = (_('extend_volume, '
                     'volumename: %(volumename)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'PoolType: %(pooltype)s.')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc,
                      'pooltype': CONSTANTS.POOL_TYPE_dic[pooltype]})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('extend_volume, '
                  'volumename: %(volumename)s, '
                  'Return code: %(rc)lu, '
                  'Error: %(errordesc)s, '
                  'Pool Name: %(eternus_pool)s, '
                  'Pool Type: %(pooltype)s.',
                  {'volumename': volumename,
                   'rc': rc,
                   'errordesc': errordesc,
                   'eternus_pool': pool_name,
                   'pooltype': CONSTANTS.POOL_TYPE_dic[pooltype]})

        return pool_name

    @lockutils.synchronized('ETERNUS-update', 'cinder-', True)
    def update_volume_stats(self):
        """Get pool capacity."""

        self.conn = self._get_eternus_connection()

        poolname_list = self._get_drvcfg('EternusPool', multiple=True)
        self._find_pools(poolname_list, self.conn)

        return (self.stats, poolname_list)

    def _get_mapdata(self, vol_instance, connector, target_portlist):
        """return mapping information."""
        mapdata = None
        multipath = connector.get('multipath', False)

        LOG.debug('_get_mapdata, volume name: %(vname)s, '
                  'protocol: %(prtcl)s, multipath: %(mpath)s.',
                  {'vname': vol_instance['ElementName'],
                   'prtcl': self.protocol, 'mpath': multipath})

        # find affinity group
        # attach the connector and include the volume
        aglist = self._find_affinity_group(connector, vol_instance)
        if not aglist:
            LOG.debug('_get_mapdata, ag_list:%s.', aglist)
        else:
            if self.protocol == 'fc':
                mapdata = self._get_mapdata_fc(aglist, vol_instance,
                                               target_portlist)
            elif self.protocol == 'iSCSI':
                mapdata = self._get_mapdata_iscsi(aglist, vol_instance,
                                                  multipath)

        LOG.debug('_get_mapdata, mapdata: %s.', mapdata)
        return mapdata

    def _get_mapdata_fc(self, aglist, vol_instance, target_portlist):
        """_get_mapdata for FibreChannel."""
        target_wwn = []

        try:
            ag_volmaplist = self._reference_eternus_names(
                aglist[0],
                ResultClass='CIM_ProtocolControllerForUnit')
            vo_volmaplist = self._reference_eternus_names(
                vol_instance.path,
                ResultClass='CIM_ProtocolControllerForUnit')
        except pywbem.CIM_Error:
            msg = (_('_get_mapdata_fc, '
                     'getting host-affinity from aglist/vol_instance failed, '
                     'affinitygroup: %(ag)s, '
                     'ReferenceNames, '
                     'cannot connect to ETERNUS.')
                   % {'ag': aglist[0]})
            LOG.exception(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        volmap = None
        for vo_volmap in vo_volmaplist:
            if vo_volmap in ag_volmaplist:
                volmap = vo_volmap
                break

        try:
            volmapinstance = self._get_eternus_instance(
                volmap,
                LocalOnly=False)
        except pywbem.CIM_Error:
            msg = (_('_get_mapdata_fc, '
                     'getting host-affinity instance failed, '
                     'volmap: %(volmap)s, '
                     'GetInstance, '
                     'cannot connect to ETERNUS.')
                   % {'volmap': volmap})
            LOG.exception(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        target_lun = int(volmapinstance['DeviceNumber'], 16)

        for target_port in target_portlist:
            target_wwn.append(target_port['Name'])

        mapdata = {'target_wwn': target_wwn,
                   'target_lun': target_lun}
        LOG.debug('_get_mapdata_fc, mapdata: %s.', mapdata)
        return mapdata

    def _get_mapdata_iscsi(self, aglist, vol_instance, multipath):
        """_get_mapdata for iSCSI."""
        target_portals = []
        target_iqns = []
        target_luns = []

        try:
            vo_volmaplist = self._reference_eternus_names(
                vol_instance.path,
                ResultClass='CIM_ProtocolControllerForUnit')
        except Exception:
            msg = (_('_get_mapdata_iscsi, '
                     'vol_instance: %(vol_instance)s, '
                     'ReferenceNames: CIM_ProtocolControllerForUnit, '
                     'cannot connect to ETERNUS.')
                   % {'vol_instance': vol_instance})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        target_properties_list = self._get_eternus_iscsi_properties()
        target_list = [prop[0] for prop in target_properties_list]
        properties_list = (
            [(prop[1], prop[2]) for prop in target_properties_list])

        for ag in aglist:
            try:
                iscsi_endpointlist = (
                    self._assoc_eternus_names(
                        ag,
                        AssocClass='FUJITSU_SAPAvailableForElement',
                        ResultClass='FUJITSU_iSCSIProtocolEndpoint'))
            except Exception:
                msg = (_('_get_mapdata_iscsi, '
                         'Associators: FUJITSU_SAPAvailableForElement, '
                         'cannot connect to ETERNUS.'))
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            iscsi_endpoint = iscsi_endpointlist[0]
            if iscsi_endpoint not in target_list:
                continue

            idx = target_list.index(iscsi_endpoint)
            target_portal, target_iqn = properties_list[idx]

            try:
                ag_volmaplist = self._reference_eternus_names(
                    ag,
                    ResultClass='CIM_ProtocolControllerForUnit')
            except Exception:
                msg = (_('_get_mapdata_iscsi, '
                         'affinitygroup: %(ag)s, '
                         'ReferenceNames, '
                         'cannot connect to ETERNUS.')
                       % {'ag': ag})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            volmap = None
            for vo_volmap in vo_volmaplist:
                if vo_volmap in ag_volmaplist:
                    volmap = vo_volmap
                    break

            if volmap is None:
                continue

            try:
                volmapinstance = self._get_eternus_instance(
                    volmap,
                    LocalOnly=False)
            except Exception:
                msg = (_('_get_mapdata_iscsi, '
                         'volmap: %(volmap)s, '
                         'GetInstance, '
                         'cannot connect to ETERNUS.')
                       % {'volmap': volmap})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            target_lun = int(volmapinstance['DeviceNumber'], 16)

            target_portals.append(target_portal)
            target_iqns.append(target_iqn)
            target_luns.append(target_lun)

        if multipath:
            mapdata = {'target_portals': target_portals,
                       'target_iqns': target_iqns,
                       'target_luns': target_luns}
        else:
            mapdata = {'target_portal': target_portals[0],
                       'target_iqn': target_iqns[0],
                       'target_lun': target_luns[0]}

        LOG.debug('_get_mapdata_iscsi, mapdata: %s.', mapdata)
        return mapdata

    def _get_drvcfg(self, tagname, filename=None, multiple=False):
        """Read from driver configuration file."""
        if not filename:
            # Set default configuration file name.
            filename = self.configuration.cinder_eternus_config_file

        LOG.debug("_get_drvcfg, input[%(filename)s][%(tagname)s].",
                  {'filename': filename, 'tagname': tagname})

        tree = ET.parse(filename)
        elem = tree.getroot()

        if not multiple:
            ret = elem.findtext(".//" + tagname)
        else:
            ret = []
            for e in elem.findall(".//" + tagname):
                if (e.text is not None) and (e.text not in ret):
                    ret.append(e.text)

        if not ret:
            msg = (_('_get_drvcfg, '
                     'filename: %(filename)s, '
                     'tagname: %(tagname)s, '
                     'data is None!! '
                     'Please edit driver configuration file and correct.')
                   % {'filename': filename,
                      'tagname': tagname})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return ret

    def _get_eternus_connection(self, filename=None):
        """return WBEM connection."""
        LOG.debug('_get_eternus_connection, filename: %s.', filename)

        ip = self._get_drvcfg('EternusIP', filename)
        port = self._get_drvcfg('EternusPort', filename)
        user = self._get_drvcfg('EternusUser', filename)
        passwd = self._get_drvcfg('EternusPassword', filename)
        url = 'http://' + ip + ':' + port

        conn = pywbem.WBEMConnection(url, (user, passwd),
                                     default_namespace='root/eternus')

        if conn is None:
            msg = (_('_get_eternus_connection, '
                     'filename: %(filename)s, '
                     'ip: %(ip)s, '
                     'port: %(port)s, '
                     'user: %(user)s, '
                     'passwd: ****, '
                     'url: %(url)s, '
                     'FAILED!!.')
                   % {'filename': filename,
                      'ip': ip,
                      'port': port,
                      'user': user,
                      'url': url})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_get_eternus_connection, conn: %s.', conn)
        return conn

    def _create_volume_name(self, id_code):
        """create volume_name on ETERNUS from id on OpenStack."""
        LOG.debug('_create_volume_name, id_code: %s.', id_code)

        if id_code is None:
            msg = _('_create_volume_name, id_code is None.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        m = md5(usedforsecurity=False)
        m.update(id_code.encode('utf-8'))

        # pylint: disable=E1121
        volumename = base64.urlsafe_b64encode(m.digest()).decode()
        ret = CONSTANTS.VOL_PREFIX + six.text_type(volumename)

        LOG.debug('_create_volume_name, ret: %s', ret)
        return ret

    def _find_pool(self, eternus_pool, detail=False):
        """find Instance or InstanceName of pool by pool name on ETERNUS."""
        LOG.debug('_find_pool, pool name: %s.', eternus_pool)

        tppoollist = []
        rgpoollist = []

        # Get pools info form CIM instance(include info about instance path).
        try:
            tppoollist = self._enum_eternus_instances(
                'FUJITSU_ThinProvisioningPool')
            rgpoollist = self._enum_eternus_instances(
                'FUJITSU_RAIDStoragePool')
        except Exception:
            msg = (_('_find_pool, '
                     'eternus_pool:%(eternus_pool)s, '
                     'EnumerateInstances, '
                     'cannot connect to ETERNUS.')
                   % {'eternus_pool': eternus_pool})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Make total pools list.
        poollist = tppoollist + rgpoollist

        # One eternus backend has only one special pool name
        # so just use pool name can get the target pool.
        for pool in poollist:
            if pool['ElementName'] == eternus_pool:
                poolinstance = pool
                break
        else:
            poolinstance = None

        if poolinstance is None:
            ret = None
        elif detail is True:
            ret = poolinstance
        else:
            ret = poolinstance.path

        LOG.debug('_find_pool, pool: %s.', ret)
        return ret

    def _find_pools(self, poolname_list, conn):
        """Find Instance or InstanceName of pool by pool name on ETERNUS."""
        LOG.debug('_find_pool, pool name: %s.', poolname_list)

        target_poolname = list(poolname_list)
        pools = []

        # Get pools info from CIM instance(include info about instance path).
        try:
            tppoollist = self._enum_eternus_instances(
                'FUJITSU_ThinProvisioningPool', conn=conn)
            rgpoollist = self._enum_eternus_instances(
                'FUJITSU_RAIDStoragePool', conn=conn)
        except Exception:
            msg = (_('_find_pool, '
                     'eternus_pool:%(eternus_pool)s, '
                     'EnumerateInstances, '
                     'cannot connect to ETERNUS.')
                   % {'eternus_pool': target_poolname})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Make total pools list.
        tppools = [(tppool, 'TPP') for tppool in tppoollist]
        rgpools = [(rgpool, 'RAID') for rgpool in rgpoollist]
        poollist = tppools + rgpools

        # One eternus backend has only one special pool name
        # so just use pool name can get the target pool.
        for pool, ptype in poollist:
            poolname = pool['ElementName']

            LOG.debug('_find_pools, '
                      'pool: %(pool)s, ptype: %(ptype)s.',
                      {'pool': poolname, 'ptype': ptype})
            if poolname in target_poolname:
                try:
                    volume_list = self._assoc_eternus_names(
                        pool.path,
                        conn=conn,
                        AssocClass='FUJITSU_AllocatedFromStoragePool',
                        ResultClass='FUJITSU_StorageVolume')

                    volume_count = len(volume_list)
                except Exception:
                    msg = (_('_find_pools, '
                             'poolname: %(poolname)s, '
                             'pooltype: %(ptype)s, '
                             'Associator Names, '
                             'cannot connect to ETERNUS.')
                           % {'ptype': ptype,
                              'poolname': poolname})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                if ptype == 'TPP':
                    param_dict = {
                        'pool-name': poolname
                    }
                    rc, errordesc, data = self._exec_eternus_cli(
                        'show_pool_provision', **param_dict)

                    if rc != 0:
                        msg = (_('_find_pools, show_pool_provision, '
                                 'pool name: %(pool_name)s, '
                                 'Return code: %(rc)lu, '
                                 'Error: %(errordesc)s, '
                                 'Message: %(job)s.')
                               % {'pool_name': poolname,
                                  'rc': rc,
                                  'errordesc': errordesc,
                                  'job': data})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

                    pool.provisioned_capacity_gb = data

                poolinfo = self.create_pool_info(pool, volume_count, ptype)

                target_poolname.remove(poolname)
                pools.append((poolinfo, poolname))

            if not target_poolname:
                break

        if not pools:
            LOG.warning('_find_pools, all the EternusPools in driver '
                        'configuration file do not exist. '
                        'Please edit the driver configuration file '
                        'to include EternusPool names.')

        # Sort pools in the order defined in driver configuration file.
        sorted_pools = (
            [pool for name in poolname_list for pool, pname in pools
             if name == pname])

        LOG.debug('_find_pools, '
                  'pools: %(pools)s, '
                  'notfound_pools: %(notfound_pools)s.',
                  {'pools': pools,
                   'notfound_pools': target_poolname})
        pools_stats = {'pools': []}
        for pool in sorted_pools:
            single_pool = {}
            if pool['type'] == 'TPP':
                thin_enabled = True
                max_ratio = self.configuration.max_over_subscription_ratio
            else:
                thin_enabled = False
                max_ratio = 1

            single_pool.update(dict(
                path=pool['path'],
                pool_name=pool['name'],
                total_capacity_gb=pool['total_capacity_gb'],
                total_volumes=pool['volume_count'],
                free_capacity_gb=pool['free_capacity_gb'],
                provisioned_capacity_gb=pool['provisioned_capacity_gb'],
                useable_capacity_gb=pool['useable_capacity_gb'],
                thin_provisioning_support=thin_enabled,
                thick_provisioning_support=not thin_enabled,
                max_over_subscription_ratio=max_ratio,
            ))
            single_pool['multiattach'] = False
            pools_stats['pools'].append(single_pool)

        self.stats['shared_targets'] = True
        self.stats['backend_state'] = 'up'
        self.stats['pools'] = pools_stats['pools']

        return self.stats, target_poolname

    def _find_eternus_service(self, classname):
        """find CIM instance about service information."""
        LOG.debug('_find_eternus_service, '
                  'classname: %s.', classname)

        try:
            services = self._enum_eternus_instance_names(
                six.text_type(classname))
        except Exception:
            msg = (_('_find_eternus_service, '
                     'classname: %(classname)s, '
                     'EnumerateInstanceNames, '
                     'cannot connect to ETERNUS.')
                   % {'classname': classname})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        ret = services[0]
        LOG.debug('_find_eternus_service, '
                  'classname: %(classname)s, '
                  'ret: %(ret)s.',
                  {'classname': classname, 'ret': ret})
        return ret

    @lockutils.synchronized('ETERNUS-SMIS-exec', 'cinder-', True)
    @utils.retry(exception.VolumeBackendAPIException)
    def _exec_eternus_service(self, classname, instanceNameList, **param_dict):
        """Execute SMI-S Method."""
        LOG.debug('_exec_eternus_service, '
                  'classname: %(a)s, '
                  'instanceNameList: %(b)s, '
                  'parameters: %(c)s.',
                  {'a': classname,
                   'b': instanceNameList,
                   'c': param_dict})
        rc = None
        retdata = None
        # Use InvokeMethod.
        try:
            rc, retdata = self.conn.InvokeMethod(
                classname,
                instanceNameList,
                **param_dict)
        except Exception:
            if rc is None:
                msg = (_('_exec_eternus_service, '
                         'classname: %(classname)s, '
                         'InvokeMethod, '
                         'cannot connect to ETERNUS.')
                       % {'classname': classname})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        # If the result has job information, wait for job complete
        if "Job" in retdata:
            rc = self._wait_for_job_complete(self.conn, retdata)

        if rc == CONSTANTS.DEVICE_IS_BUSY:
            msg = _('Device is in Busy state')
            raise exception.VolumeBackendAPIException(data=msg)

        errordesc = CONSTANTS.RETCODE_dic.get(six.text_type(rc),
                                              CONSTANTS.UNDEF_MSG)

        ret = (rc, errordesc, retdata)

        LOG.debug('_exec_eternus_service, '
                  'classname: %(a)s, '
                  'instanceNameList: %(b)s, '
                  'parameters: %(c)s, '
                  'Return code: %(rc)s, '
                  'Error: %(errordesc)s, '
                  'Return data: %(retdata)s.',
                  {'a': classname,
                   'b': instanceNameList,
                   'c': param_dict,
                   'rc': rc,
                   'errordesc': errordesc,
                   'retdata': retdata})
        return ret

    @lockutils.synchronized('ETERNUS-SMIS-other', 'cinder-', True)
    @utils.retry(exception.VolumeBackendAPIException)
    def _enum_eternus_instances(self, classname, conn=None, **param_dict):
        """Enumerate Instances."""
        LOG.debug('_enum_eternus_instances, classname: %s.', classname)

        if not conn:
            conn = self.conn

        ret = conn.EnumerateInstances(classname, **param_dict)

        LOG.debug('_enum_eternus_instances, enum %d instances.', len(ret))
        return ret

    @lockutils.synchronized('ETERNUS-SMIS-other', 'cinder-', True)
    @utils.retry(exception.VolumeBackendAPIException)
    def _enum_eternus_instance_names(self, classname):
        """Enumerate Instance Names."""
        LOG.debug('_enum_eternus_instance_names, classname: %s.', classname)

        ret = self.conn.EnumerateInstanceNames(classname)

        LOG.debug('_enum_eternus_instance_names, enum %d names.', len(ret))
        return ret

    @lockutils.synchronized('ETERNUS-SMIS-getinstance', 'cinder-', True)
    @utils.retry(exception.VolumeBackendAPIException)
    def _get_eternus_instance(self, classname, **param_dict):
        """Get Instance."""
        LOG.debug('_get_eternus_instance, '
                  'classname: %(cls)s, param: %(param)s.',
                  {'cls': classname, 'param': param_dict})

        ret = self.conn.GetInstance(classname, **param_dict)

        LOG.debug('_get_eternus_instance, ret: %s.', ret)
        return ret

    @lockutils.synchronized('ETERNUS-SMIS-other', 'cinder-', True)
    @utils.retry(exception.VolumeBackendAPIException)
    def _assoc_eternus(self, classname, conn=None, **param_dict):
        """Associator."""
        LOG.debug('_assoc_eternus, '
                  'classname: %(cls)s, param: %(param)s.',
                  {'cls': classname, 'param': param_dict})

        if not conn:
            conn = self.conn

        ret = conn.Associators(classname, **param_dict)

        LOG.debug('_assoc_eternus, enum %d instances.', len(ret))
        return ret

    @lockutils.synchronized('ETERNUS-SMIS-other', 'cinder-', True)
    @utils.retry(exception.VolumeBackendAPIException)
    def _assoc_eternus_names(self, classname, conn=None, **param_dict):
        """Associator Names."""
        LOG.debug('_assoc_eternus_names, '
                  'classname: %(cls)s, param: %(param)s.',
                  {'cls': classname, 'param': param_dict})

        if not conn:
            conn = self.conn

        ret = conn.AssociatorNames(classname, **param_dict)

        LOG.debug('_assoc_eternus_names, enum %d names.', len(ret))
        return ret

    @lockutils.synchronized('ETERNUS-SMIS-other', 'cinder-', True)
    @utils.retry(exception.VolumeBackendAPIException)
    def _reference_eternus_names(self, classname, **param_dict):
        """Refference Names."""
        LOG.debug('_reference_eternus_names, '
                  'classname: %(cls)s, param: %(param)s.',
                  {'cls': classname, 'param': param_dict})

        ret = self.conn.ReferenceNames(classname, **param_dict)

        LOG.debug('_reference_eternus_names, enum %d names.', len(ret))
        return ret

    def _create_eternus_instance_name(self, classname, bindings):
        """create CIM InstanceName from classname and bindings."""
        LOG.debug('_create_eternus_instance_name, '
                  'classname: %(cls)s, bindings: %(bind)s.',
                  {'cls': classname, 'bind': bindings})

        instancename = None

        try:
            instancename = pywbem.CIMInstanceName(
                classname,
                namespace='root/eternus',
                keybindings=bindings)
        except NameError:
            instancename = None

        LOG.debug('_create_eternus_instance_name, ret: %s.', instancename)
        return instancename

    def _find_lun(self, volume):
        """find lun instance from volume class or volumename on ETERNUS."""
        LOG.debug('_find_lun, volume id: %s.', volume['id'])
        volumeinstance = None
        volumename = self._create_volume_name(volume['id'])

        try:
            location = ast.literal_eval(volume['provider_location'])
            classname = location['classname']
            bindings = location['keybindings']

            if classname and bindings:
                LOG.debug('_find_lun, '
                          'classname: %(classname)s, '
                          'bindings: %(bindings)s.',
                          {'classname': classname,
                           'bindings': bindings})
                volume_instance_name = (
                    self._create_eternus_instance_name(classname, bindings))

                LOG.debug('_find_lun, '
                          'volume_insatnce_name: %(volume_instance_name)s.',
                          {'volume_instance_name': volume_instance_name})

                vol_instance = (
                    self._get_eternus_instance(volume_instance_name))

                if vol_instance['ElementName'] == volumename:
                    volumeinstance = vol_instance
        except Exception:
            volumeinstance = None
            LOG.debug('_find_lun, '
                      'Cannot get volume instance from provider location, '
                      'Search all volume using EnumerateInstanceNames.')

        if volumeinstance is None:
            # for old version

            LOG.debug('_find_lun, '
                      'volumename: %(volumename)s.',
                      {'volumename': volumename})

            # get volume instance from volumename on ETERNUS
            try:
                namelist = self._enum_eternus_instance_names(
                    'FUJITSU_StorageVolume')
            except Exception:
                msg = (_('_find_lun, '
                         'volumename: %(volumename)s, '
                         'EnumerateInstanceNames, '
                         'cannot connect to ETERNUS.')
                       % {'volumename': volumename})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            for name in namelist:
                try:
                    vol_instance = self._get_eternus_instance(name)

                    if vol_instance['ElementName'] == volumename:
                        volumeinstance = vol_instance
                        path = volumeinstance.path

                        LOG.debug('_find_lun, '
                                  'volumename: %(volumename)s, '
                                  'vol_instance: %(vol_instance)s.',
                                  {'volumename': volumename,
                                   'vol_instance': path})
                        break
                except Exception:
                    continue
            else:
                LOG.debug('_find_lun, '
                          'volumename: %(volumename)s, '
                          'volume not found on ETERNUS.',
                          {'volumename': volumename})

        LOG.debug('_find_lun, ret: %s.', volumeinstance)
        return volumeinstance

    def _find_copysession(self, vol_instance):
        """find copysession from volumename on ETERNUS."""
        LOG.debug('_find_copysession, volume name: %s.',
                  vol_instance['ElementName'])

        try:
            cpsessionlist = self.conn.ReferenceNames(
                vol_instance.path,
                ResultClass='FUJITSU_StorageSynchronized')
        except Exception:
            msg = (_('_find_copysession, '
                     'ReferenceNames, '
                     'vol_instance: %(vol_instance_path)s, '
                     'Cannot connect to ETERNUS.')
                   % {'vol_instance_path': vol_instance.path})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_copysession, '
                  'cpsessionlist: %(cpsessionlist)s.',
                  {'cpsessionlist': cpsessionlist})

        LOG.debug('_find_copysession, ret: %s.', cpsessionlist)
        return cpsessionlist

    def _wait_for_copy_complete(self, cpsession):
        """Wait for the completion of copy."""
        LOG.debug('_wait_for_copy_complete, cpsession: %s.', cpsession)

        cpsession_instance = None

        while True:
            try:
                cpsession_instance = self.conn.GetInstance(
                    cpsession,
                    LocalOnly=False)
            except Exception:
                cpsession_instance = None

            # if copy session is none,
            # it means copy session was finished,break and return
            if cpsession_instance is None:
                break

            LOG.debug('_wait_for_copy_complete, '
                      'find target copysession, '
                      'wait for end of copysession.')

            if cpsession_instance['CopyState'] == CONSTANTS.BROKEN:
                msg = (_('_wait_for_copy_complete, '
                         'cpsession: %(cpsession)s, '
                         'copysession state is BROKEN.')
                       % {'cpsession': cpsession})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            time.sleep(10)

    @utils.retry(exception.VolumeBackendAPIException)
    def _delete_copysession(self, cpsession):
        """delete copysession."""
        LOG.debug('_delete_copysession: cpssession: %s.', cpsession)

        try:
            cpsession_instance = self._get_eternus_instance(
                cpsession, LocalOnly=False)
        except Exception:
            LOG.info('_delete_copysession, '
                     'the copysession was already completed.')
            return

        copytype = cpsession_instance['CopyType']

        # set oparation code
        # SnapOPC: 19 (Return To ResourcePool)
        # OPC:8 (Detach)
        # EC/REC:8 (Detach)
        operation = CONSTANTS.OPERATION_dic.get(copytype, None)
        if operation is None:
            msg = (_('_delete_copysession, '
                     'copy session type is undefined! '
                     'copy session: %(cpsession)s, '
                     'copy type: %(copytype)s.')
                   % {'cpsession': cpsession,
                      'copytype': copytype})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        repservice = self._find_eternus_service(CONSTANTS.REPL)
        if repservice is None:
            msg = (_('_delete_copysession, '
                     'Cannot find Replication Service'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Invoke method for delete copysession
        rc, errordesc, job = self._exec_eternus_service(
            'ModifyReplicaSynchronization',
            repservice,
            Operation=self._pywbem_uint(operation, '16'),
            Synchronization=cpsession,
            Force=True,
            WaitForCopyState=self._pywbem_uint(15, '16'))

        LOG.debug('_delete_copysession, '
                  'copysession: %(cpsession)s, '
                  'operation: %(operation)s, '
                  'Return code: %(rc)lu, '
                  'Error: %(errordesc)s.',
                  {'cpsession': cpsession,
                   'operation': operation,
                   'rc': rc,
                   'errordesc': errordesc})

        if rc == CONSTANTS.COPYSESSION_NOT_EXIST:
            LOG.debug('_delete_copysession, '
                      'cpsession: %(cpsession)s, '
                      'copysession is not exist.',
                      {'cpsession': cpsession})
        elif rc == CONSTANTS.VOLUME_IS_BUSY:
            msg = (_('_delete_copysession, '
                     'copysession: %(cpsession)s, '
                     'operation: %(operation)s, '
                     'Error: Volume is in Busy state')
                   % {'cpsession': cpsession,
                      'operation': operation})
            raise exception.VolumeIsBusy(msg)
        elif rc != 0:
            msg = (_('_delete_copysession, '
                     'copysession: %(cpsession)s, '
                     'operation: %(operation)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s.')
                   % {'cpsession': cpsession,
                      'operation': operation,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def _get_target_port(self):
        """return target portid."""
        LOG.debug('_get_target_port, protocol: %s.', self.protocol)

        target_portlist = []
        if self.protocol == 'fc':
            prtcl_endpoint = 'FUJITSU_SCSIProtocolEndpoint'
            connection_type = 2
        elif self.protocol == 'iSCSI':
            prtcl_endpoint = 'FUJITSU_iSCSIProtocolEndpoint'
            connection_type = 7

        try:
            tgtportlist = self._enum_eternus_instances(prtcl_endpoint)
        except Exception:
            msg = (_('_get_target_port, '
                     'EnumerateInstances, '
                     'cannot connect to ETERNUS.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for tgtport in tgtportlist:
            # Check : protocol of tgtport
            if tgtport['ConnectionType'] != connection_type:
                continue

            # Check : if port is for remote copy, continue
            if (tgtport['RAMode'] & 0x7B) != 0x00:
                continue

            # Check : if port is for StorageCluster, continue
            if 'SCGroupNo' in tgtport:
                continue

            target_portlist.append(tgtport)

            LOG.debug('_get_target_port, '
                      'connection type: %(cont)s, '
                      'ramode: %(ramode)s.',
                      {'cont': tgtport['ConnectionType'],
                       'ramode': tgtport['RAMode']})

        LOG.debug('_get_target_port, '
                  'target port: %(target_portid)s.',
                  {'target_portid': target_portlist})

        if len(target_portlist) == 0:
            msg = (_('_get_target_port, '
                     'protcol: %(protocol)s, '
                     'target_port not found.')
                   % {'protocol': self.protocol})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_get_target_port, ret: %s.', target_portlist)
        return target_portlist

    @lockutils.synchronized('ETERNUS-connect', 'cinder-', True)
    def _map_lun(self, vol_instance, connector, targetlist=None):
        """map volume to host."""
        volumename = vol_instance['ElementName']
        LOG.debug('_map_lun, '
                  'volume name: %(vname)s, connector: %(connector)s.',
                  {'vname': volumename, 'connector': connector})

        volume_uid = vol_instance['Name']
        initiatorlist = self._find_initiator_names(connector)
        aglist = self._find_affinity_group(connector)
        configservice = self._find_eternus_service(CONSTANTS.CTRL_CONF)

        if targetlist is None:
            targetlist = self._get_target_port()

        if configservice is None:
            msg = (_('_map_lun, '
                     'vol_instance.path:%(vol)s, '
                     'volumename: %(volumename)s, '
                     'volume_uid: %(uid)s, '
                     'initiator: %(initiator)s, '
                     'target: %(tgt)s, '
                     'aglist: %(aglist)s, '
                     'Storage Configuration Service not found.')
                   % {'vol': vol_instance.path,
                      'volumename': volumename,
                      'uid': volume_uid,
                      'initiator': initiatorlist,
                      'tgt': targetlist,
                      'aglist': aglist})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_map_lun, '
                  'vol_instance.path: %(vol_instance)s, '
                  'volumename:%(volumename)s, '
                  'initiator:%(initiator)s, '
                  'target:%(tgt)s.',
                  {'vol_instance': vol_instance.path,
                   'volumename': [volumename],
                   'initiator': initiatorlist,
                   'tgt': targetlist})

        if not aglist:
            # Create affinity group and set host-affinity.
            for target in targetlist:
                LOG.debug('_map_lun, '
                          'lun_name: %(volume_uid)s, '
                          'Initiator: %(initiator)s, '
                          'target: %(target)s.',
                          {'volume_uid': [volume_uid],
                           'initiator': initiatorlist,
                           'target': target['Name']})

                rc, errordesc, job = self._exec_eternus_service(
                    'ExposePaths',
                    configservice,
                    LUNames=[volume_uid],
                    InitiatorPortIDs=initiatorlist,
                    TargetPortIDs=[target['Name']],
                    DeviceAccesses=[self._pywbem_uint(2, '16')])

                LOG.debug('_map_lun, '
                          'Error: %(errordesc)s, '
                          'Return code: %(rc)lu, '
                          'Create affinitygroup and set host-affinity.',
                          {'errordesc': errordesc,
                           'rc': rc})

                if rc != 0 and rc != CONSTANTS.LUNAME_IN_USE:
                    LOG.warning('_map_lun, '
                                'lun_name: %(volume_uid)s, '
                                'Initiator: %(initiator)s, '
                                'target: %(target)s, '
                                'Return code: %(rc)lu, '
                                'Error: %(errordesc)s.',
                                {'volume_uid': [volume_uid],
                                 'initiator': initiatorlist,
                                 'target': target['Name'],
                                 'rc': rc,
                                 'errordesc': errordesc})
        else:
            # Add lun to affinity group
            for ag in aglist:
                LOG.debug('_map_lun, '
                          'ag: %(ag)s, lun_name: %(volume_uid)s.',
                          {'ag': ag,
                           'volume_uid': volume_uid})

                rc, errordesc, job = self._exec_eternus_service(
                    'ExposePaths',
                    configservice, LUNames=[volume_uid],
                    DeviceAccesses=[self._pywbem_uint(2, '16')],
                    ProtocolControllers=[ag])

                LOG.debug('_map_lun, '
                          'Error: %(errordesc)s, '
                          'Return code: %(rc)lu, '
                          'Add lun to affinity group.',
                          {'errordesc': errordesc,
                           'rc': rc})

                if rc != 0 and rc != CONSTANTS.LUNAME_IN_USE:
                    LOG.warning('_map_lun, '
                                'lun_name: %(volume_uid)s, '
                                'Initiator: %(initiator)s, '
                                'ag: %(ag)s, '
                                'Return code: %(rc)lu, '
                                'Error: %(errordesc)s.',
                                {'volume_uid': [volume_uid],
                                 'initiator': initiatorlist,
                                 'ag': ag,
                                 'rc': rc,
                                 'errordesc': errordesc})

    def _find_initiator_names(self, connector):
        """return initiator names."""

        initiatornamelist = []

        if self.protocol == 'fc' and connector['wwpns']:
            LOG.debug('_find_initiator_names, wwpns: %s.',
                      connector['wwpns'])
            initiatornamelist = connector['wwpns']
        elif self.protocol == 'iSCSI' and connector['initiator']:
            LOG.debug('_find_initiator_names, initiator: %s.',
                      connector['initiator'])
            initiatornamelist.append(connector['initiator'])

        if not initiatornamelist:
            msg = (_('_find_initiator_names, '
                     'connector: %(connector)s, '
                     'initiator not found.')
                   % {'connector': connector})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_initiator_names, '
                  'initiator list: %(initiator)s.',
                  {'initiator': initiatornamelist})

        return initiatornamelist

    def _find_affinity_group(self, connector, vol_instance=None):
        """find affinity group from connector."""
        LOG.debug('_find_affinity_group, vol_instance: %s.', vol_instance)

        affinity_grouplist = []
        initiatorlist = self._find_initiator_names(connector)

        if vol_instance is None:
            try:
                aglist = self._enum_eternus_instance_names(
                    'FUJITSU_AffinityGroupController')
            except Exception:
                msg = (_('_find_affinity_group, '
                         'connector: %(connector)s, '
                         'EnumerateInstanceNames, '
                         'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_find_affinity_group,'
                      'affinity_groups:%s', aglist)
        else:
            try:
                aglist = self._assoc_eternus_names(
                    vol_instance.path,
                    AssocClass='FUJITSU_ProtocolControllerForUnit',
                    ResultClass='FUJITSU_AffinityGroupController')
            except Exception:
                msg = (_('_find_affinity_group,'
                         'connector: %(connector)s,'
                         'AssocNames: FUJITSU_ProtocolControllerForUnit, '
                         'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_find_affinity_group, '
                      'vol_instance.path: %(volume)s, '
                      'affinity_groups: %(aglist)s.',
                      {'volume': vol_instance.path,
                       'aglist': aglist})

        for ag in aglist:
            try:
                hostaglist = self._assoc_eternus(
                    ag,
                    AssocClass='FUJITSU_AuthorizedTarget',
                    ResultClass='FUJITSU_AuthorizedPrivilege')
            except Exception:
                msg = (_('_find_affinity_group, '
                         'connector: %(connector)s, '
                         'Associators: FUJITSU_AuthorizedTarget, '
                         'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            for hostag in hostaglist:
                for initiator in initiatorlist:
                    if initiator.lower() not in hostag['InstanceID'].lower():
                        continue

                    LOG.debug('_find_affinity_group, '
                              'AffinityGroup: %(ag)s.', {'ag': ag})
                    affinity_grouplist.append(ag)
                    break
                break

        LOG.debug('_find_affinity_group, '
                  'initiators: %(initiator)s, '
                  'affinity_group: %(affinity_group)s.',
                  {'initiator': initiatorlist,
                   'affinity_group': affinity_grouplist})
        return affinity_grouplist

    @lockutils.synchronized('ETERNUS-connect', 'cinder-', True)
    def _unmap_lun(self, volume, connector, force=False):
        """unmap volume from host."""
        LOG.debug('_map_lun, volume id: %(vid)s, '
                  'connector: %(connector)s, force: %(frc)s.',
                  {'vid': volume['id'],
                   'connector': connector, 'frc': force})

        volumename = self._create_volume_name(volume['id'])
        vol_instance = self._find_lun(volume)
        if vol_instance is None:
            LOG.info('_unmap_lun, '
                     'volumename:%(volumename)s, '
                     'volume not found.',
                     {'volumename': volumename})
            return False

        volume_uid = vol_instance['Name']

        if not force:
            aglist = self._find_affinity_group(connector, vol_instance)
            if not aglist:
                LOG.info('_unmap_lun, '
                         'volumename: %(volumename)s, '
                         'volume is not mapped.',
                         {'volumename': volumename})
                return False
        else:
            try:
                aglist = self._assoc_eternus_names(
                    vol_instance.path,
                    AssocClass='CIM_ProtocolControllerForUnit',
                    ResultClass='FUJITSU_AffinityGroupController')
            except Exception:
                msg = (_('_unmap_lun,'
                         'vol_instance.path: %(volume)s, '
                         'AssociatorNames: CIM_ProtocolControllerForUnit, '
                         'cannot connect to ETERNUS.')
                       % {'volume': vol_instance.path})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_unmap_lun, '
                      'vol_instance.path: %(volume)s, '
                      'affinity_groups: %(aglist)s.',
                      {'volume': vol_instance.path,
                       'aglist': aglist})

        configservice = self._find_eternus_service(CONSTANTS.CTRL_CONF)
        if configservice is None:
            msg = (_('_unmap_lun, '
                     'vol_instance.path: %(volume)s, '
                     'volumename: %(volumename)s, '
                     'volume_uid: %(uid)s, '
                     'aglist: %(aglist)s, '
                     'Controller Configuration Service not found.')
                   % {'vol': vol_instance.path,
                      'volumename': [volumename],
                      'uid': [volume_uid],
                      'aglist': aglist})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for ag in aglist:
            LOG.debug('_unmap_lun, '
                      'volumename: %(volumename)s, '
                      'volume_uid: %(volume_uid)s, '
                      'AffinityGroup: %(ag)s.',
                      {'volumename': volumename,
                       'volume_uid': volume_uid,
                       'ag': ag})

            rc, errordesc, job = self._exec_eternus_service(
                'HidePaths',
                configservice,
                LUNames=[volume_uid],
                ProtocolControllers=[ag])

            LOG.debug('_unmap_lun, '
                      'Error: %(errordesc)s, '
                      'Return code: %(rc)lu.',
                      {'errordesc': errordesc,
                       'rc': rc})

            if rc == CONSTANTS.LUNAME_NOT_EXIST:
                LOG.debug('_unmap_lun, '
                          'volumename: %(volumename)s, '
                          'Invalid LUNames.',
                          {'volumename': volumename})
            elif rc != 0:
                msg = (_('_unmap_lun, '
                         'volumename: %(volumename)s, '
                         'volume_uid: %(volume_uid)s, '
                         'AffinityGroup: %(ag)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s.')
                       % {'volumename': volumename,
                          'volume_uid': volume_uid,
                          'ag': ag,
                          'rc': rc,
                          'errordesc': errordesc})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_unmap_lun, '
                  'volumename: %(volumename)s.',
                  {'volumename': volumename})
        return True

    def _get_eternus_iscsi_properties(self):
        """get target port iqns and target_portals."""

        iscsi_properties_list = []
        iscsiip_list = self._get_drvcfg('EternusISCSIIP', multiple=True)
        iscsi_port = self.configuration.target_port

        LOG.debug('_get_eternus_iscsi_properties, iplist: %s.', iscsiip_list)

        try:
            ip_endpointlist = self._enum_eternus_instance_names(
                'FUJITSU_IPProtocolEndpoint')
        except Exception:
            msg = (_('_get_eternus_iscsi_properties, '
                     'iscsiip: %(iscsiip)s, '
                     'EnumerateInstanceNames, '
                     'cannot connect to ETERNUS.')
                   % {'iscsiip': iscsiip_list})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for ip_endpoint in ip_endpointlist:
            try:
                ip_endpoint_instance = self._get_eternus_instance(
                    ip_endpoint)
                ip_address = ip_endpoint_instance['IPv4Address']
                LOG.debug('_get_eternus_iscsi_properties, '
                          'instanceip: %(ip)s, '
                          'iscsiip: %(iscsiip)s.',
                          {'ip': ip_address,
                           'iscsiip': iscsiip_list})
            except Exception:
                msg = (_('_get_eternus_iscsi_properties, '
                         'iscsiip: %(iscsiip)s, '
                         'GetInstance, '
                         'cannot connect to ETERNUS.')
                       % {'iscsiip': iscsiip_list})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if ip_address not in iscsiip_list:
                continue

            LOG.debug('_get_eternus_iscsi_properties, '
                      'find iscsiip: %(ip)s.', {'ip': ip_address})
            try:
                tcp_endpointlist = self._assoc_eternus_names(
                    ip_endpoint,
                    AssocClass='CIM_BindsTo',
                    ResultClass='FUJITSU_TCPProtocolEndpoint')
            except Exception:
                msg = (_('_get_eternus_iscsi_properties, '
                         'iscsiip: %(iscsiip)s, '
                         'AssociatorNames: CIM_BindsTo, '
                         'cannot connect to ETERNUS.')
                       % {'iscsiip': ip_address})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            for tcp_endpoint in tcp_endpointlist:
                try:
                    iscsi_endpointlist = (
                        self._assoc_eternus(tcp_endpoint,
                                            AssocClass='CIM_BindsTo',
                                            ResultClass='FUJITSU_iSCSI'
                                            'ProtocolEndpoint'))
                except Exception:
                    msg = (_('_get_eternus_iscsi_properties, '
                             'iscsiip: %(iscsiip)s, '
                             'AssociatorNames: CIM_BindsTo, '
                             'cannot connect to ETERNUS.')
                           % {'iscsiip': ip_address})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                for iscsi_endpoint in iscsi_endpointlist:
                    target_portal = "%s:%s" % (ip_address, iscsi_port)
                    iqn = iscsi_endpoint['Name'].split(',')[0]
                    iscsi_properties_list.append((iscsi_endpoint.path,
                                                  target_portal,
                                                  iqn))
                    LOG.debug('_get_eternus_iscsi_properties, '
                              'target_portal: %(target_portal)s, '
                              'iqn: %(iqn)s.',
                              {'target_portal': target_portal,
                               'iqn': iqn})

        if len(iscsi_properties_list) == 0:
            msg = (_('_get_eternus_iscsi_properties, '
                     'iscsiip list: %(iscsiip_list)s, '
                     'iqn not found.')
                   % {'iscsiip_list': iscsiip_list})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        LOG.debug('_get_eternus_iscsi_properties, '
                  'iscsi_properties_list: %(iscsi_properties_list)s.',
                  {'iscsi_properties_list': iscsi_properties_list})

        return iscsi_properties_list

    def _wait_for_job_complete(self, conn, job):
        """Given the job wait for it to complete."""
        self.retries = 0
        self.wait_for_job_called = False

        def _wait_for_job_complete():
            """Called at an interval until the job is finished."""
            if self._is_job_finished(conn, job):
                raise loopingcall.LoopingCallDone()
            if self.retries > CONSTANTS.JOB_RETRIES:
                LOG.error("_wait_for_job_complete, "
                          "failed after %(retries)d tries.",
                          {'retries': self.retries})
                raise loopingcall.LoopingCallDone()

            try:
                self.retries += 1
                if not self.wait_for_job_called:
                    if self._is_job_finished(conn, job):
                        self.wait_for_job_called = True
            except Exception:
                exceptionMessage = _("Issue encountered waiting for job.")
                LOG.exception(exceptionMessage)
                raise exception.VolumeBackendAPIException(exceptionMessage)

        self.wait_for_job_called = False
        timer = loopingcall.FixedIntervalLoopingCall(_wait_for_job_complete)
        timer.start(interval=CONSTANTS.JOB_INTERVAL_SEC).wait()

        jobInstanceName = job['Job']
        jobinstance = conn.GetInstance(jobInstanceName,
                                       LocalOnly=False)

        rc = jobinstance['ErrorCode']

        LOG.debug('_wait_for_job_complete, rc: %s.', rc)
        return rc

    def _is_job_finished(self, conn, job):
        """Check if the job is finished."""
        jobInstanceName = job['Job']
        jobinstance = conn.GetInstance(jobInstanceName,
                                       LocalOnly=False)
        jobstate = jobinstance['JobState']
        LOG.debug('_is_job_finished,'
                  'state: %(state)s', {'state': jobstate})
        # From ValueMap of JobState in CIM_ConcreteJob
        # 2=New, 3=Starting, 4=Running, 32767=Queue Pending
        # ValueMap("2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13..32767,
        # 32768..65535"),
        # Values("New, Starting, Running, Suspended, Shutting Down,
        # Completed, Terminated, Killed, Exception, Service,
        # Query Pending, DMTF Reserved, Vendor Reserved")]
        # NOTE(deva): string matching based on
        #             http://ipmitool.cvs.sourceforge.net/
        #               viewvc/ipmitool/ipmitool/lib/ipmi_chassis.c

        if jobstate in [2, 3, 4]:
            job_finished = False
        else:
            job_finished = True

        LOG.debug('_is_job_finished, finish: %s.', job_finished)
        return job_finished

    @staticmethod
    def _pywbem_uint(num, datatype):
        try:
            if datatype == '8':
                result = pywbem.Uint8(num)
            elif datatype == '16':
                result = pywbem.Uint16(num)
            elif datatype == '32':
                result = pywbem.Uint32(num)
            elif datatype == '64':
                result = pywbem.Uint64(num)
        except NameError:
            result = num

        return result

    def _find_pool_from_volume(self, vol_instance, manage_type='volume'):
        """Find Instance or InstanceName of pool by volume instance."""
        LOG.debug('_find_pool_from_volume, volume: %(volume)s.',
                  {'volume': vol_instance})
        poolname = None
        target_pool = None
        filename = None
        conn = self.conn

        # Get poolname of volume on Eternus.
        try:
            pools = self._assoc_eternus(
                vol_instance.path,
                conn=conn,
                AssocClass='FUJITSU_AllocatedFromStoragePool',
                ResultClass='CIM_StoragePool')
        except Exception:
            msg = (_('_find_pool_from_volume, '
                     'vol_instance: %s, '
                     'Associators: FUJITSU_AllocatedFromStoragePool, '
                     'cannot connect to ETERNUS.')
                   % vol_instance.path)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not pools:
            msg = (_('_find_pool_from_volume, '
                     'vol_instance: %s, '
                     'pool not found.')
                   % vol_instance.path)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Get poolname from driver configuration file.
        if manage_type == 'volume':
            cfgpool_list = list(self._get_drvcfg('EternusPool',
                                                 filename=filename,
                                                 multiple=True))
        elif manage_type == 'snapshot':
            cfgpool_list = list(self._get_drvcfg('EternusSnapPool',
                                                 filename=filename,
                                                 multiple=True))
        LOG.debug('_find_pool_from_volume, cfgpool_list: %(cfgpool_list)s.',
                  {'cfgpool_list': cfgpool_list})
        for pool in pools:
            if pool['ElementName'] in cfgpool_list:
                poolname = pool['ElementName']
                target_pool = pool.path
                break

        if not target_pool:
            msg = (_('_find_pool_from_volume, '
                     'vol_instance: %s, '
                     'the pool of volume not in driver configuration file.')
                   % vol_instance.path)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_pool_from_volume, poolname: %(poolname)s, '
                  'target_pool: %(target_pool)s.',
                  {'poolname': poolname, 'target_pool': target_pool})
        return poolname, target_pool

    def _check_user(self):
        """Check whether user's role is accessible to ETERNUS and Software."""
        ret = True
        rc, errordesc, job = self._exec_eternus_cli('check_user_role')
        if rc != 0:
            msg = (_('_check_user, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(job)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if job != 'Software':
            msg = (_('_check_user, '
                     'Specified user(%(user)s) does not have '
                     'Software role: %(role)s.')
                   % {'user': self._get_drvcfg('EternusUser'),
                      'role': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return ret

    def _exec_eternus_cli(self, command, retry=CONSTANTS.TIMES_MIN,
                          retry_interval=CONSTANTS.RETRY_INTERVAL,
                          retry_code=[32787], filename=None, timeout=None,
                          **param_dict):
        """Execute ETERNUS CLI."""
        LOG.debug('_exec_eternus_cli, '
                  'command: %(a)s, '
                  'filename: %(f)s, '
                  'timeout: %(t)s, '
                  'parameters: %(b)s.',
                  {'a': command,
                   'f': filename,
                   't': timeout,
                   'b': param_dict})

        result = None
        rc = None
        retdata = None
        errordesc = None
        filename = self.configuration.cinder_eternus_config_file
        storage_ip = self._get_drvcfg('EternusIP')
        if not self.fjdxcli.get(filename):
            user = self._get_drvcfg('EternusUser')
            password = self._get_drvcfg('EternusPassword')
            self.fjdxcli[filename] = (
                eternus_dx_cli.FJDXCLI(user, storage_ip,
                                       password=password))

        for retry_num in range(retry):
            # Execute ETERNUS CLI and get return value.
            try:
                out_dict = self.fjdxcli[filename].done(command, **param_dict)
                result = out_dict.get('result')
                rc_str = out_dict.get('rc')
                retdata = out_dict.get('message')
            except Exception as ex:
                msg = (_('_exec_eternus_cli, '
                         'unexpected error: %(ex)s.')
                       % {'ex': ex})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Check ssh result.
            if result == 255:
                LOG.info('_exec_eternus_cli, retry, '
                         'command: %(command)s, '
                         'option: %(option)s, '
                         'ip: %(ip)s, '
                         'SSH Result: %(result)s, '
                         'retdata: %(retdata)s, '
                         'TryNum: %(rn)s.',
                         {'command': command,
                          'option': param_dict,
                          'ip': storage_ip,
                          'result': result,
                          'retdata': retdata,
                          'rn': (retry_num + 1)})
                time.sleep(retry_interval)
                continue
            elif result != 0:
                msg = (_('_exec_eternus_cli, '
                         'unexpected error, '
                         'command: %(command)s, '
                         'option: %(option)s, '
                         'ip: %(ip)s, '
                         'resuslt: %(result)s, '
                         'retdata: %(retdata)s.')
                       % {'command': command,
                          'option': param_dict,
                          'ip': storage_ip,
                          'result': result,
                          'retdata': retdata})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Check CLI return code.
            if rc_str.isdigit():
                # SMI-S style return code.
                rc = int(rc_str)

                try:
                    errordesc = CONSTANTS.RETCODE_dic[str(rc)]
                except Exception:
                    errordesc = 'Undefined Error!!'

                if rc in retry_code:
                    LOG.info('_exec_eternus_cli, retry, '
                             'ip: %(ip)s, '
                             'RetryCode: %(rc)s, '
                             'TryNum: %(rn)s.',
                             {'ip': storage_ip,
                              'rc': rc,
                              'rn': (retry_num + 1)})
                    time.sleep(retry_interval)
                    continue
                if rc == 4:
                    if ('Authentication failed' in retdata and
                            retry_num + 1 < retry):
                        LOG.warning('_exec_eternus_cli, retry, ip: %(ip)s, '
                                    'Message: %(message)s, '
                                    'TryNum: %(rn)s.',
                                    {'ip': storage_ip,
                                     'message': retdata,
                                     'rn': (retry_num + 1)})
                        time.sleep(1)
                        continue

                break
            else:
                # CLI style return code.
                LOG.warning('_exec_eternus_cli, '
                            'WARNING!! '
                            'ip: %(ip)s, '
                            'ReturnCode: %(rc_str)s, '
                            'ReturnData: %(retdata)s.',
                            {'ip': storage_ip,
                             'rc_str': rc_str,
                             'retdata': retdata})

                errordesc = rc_str
                rc = 4  # Failed.
                break
        else:
            if 0 < result:
                msg = (_('_exec_eternus_cli, '
                         'cannot connect to ETERNUS. '
                         'SSH Result: %(result)s, '
                         'retdata: %(retdata)s.')
                       % {'result': result,
                          'retdata': retdata})

                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                LOG.warning('_exec_eternus_cli, Retry was exceeded.')

        ret = (rc, errordesc, retdata)

        LOG.debug('_exec_eternus_cli, '
                  'command: %(a)s, '
                  'parameters: %(b)s, '
                  'ip: %(ip)s, '
                  'Return code: %(rc)s, '
                  'Error: %(errordesc)s.',
                  {'a': command,
                   'b': param_dict,
                   'ip': storage_ip,
                   'rc': rc,
                   'errordesc': errordesc})
        return ret



# Copyright 2020 toyou Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Testing for acs5000 san storage driver
"""

import copy
import json
import random
import time
from unittest import mock

from eventlet import greenthread
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils import excutils
from oslo_utils import units
import paramiko

from cinder import context
import cinder.db
from cinder import exception
from cinder import ssh_utils
from cinder.tests.unit import test
from cinder.tests.unit import utils as testutils
from cinder import utils as cinder_utils
from cinder.volume import configuration as conf
from cinder.volume.drivers.toyou.acs5000 import acs5000_common
from cinder.volume.drivers.toyou.acs5000 import acs5000_iscsi

POOLS_NAME = ['pool01', 'pool02']
VOLUME_PRE = acs5000_common.VOLUME_PREFIX
# luns number only for test
LUN_NUMS_AVAILABLE = range(0, 5)
# snapshot count on a volume, only for test
SNAPSHOTS_A_VOLUME = 3
# snapshot count on a system, only for test
SNAPSHOTS_ON_SYSTEM = 10
# volume count on a pool, only for test
VOLUME_LIMIT_ON_POOL = 10
# volume count on a pool, only for test
VOLUME_LIMIT_ON_SYSTEM = 16
# volume count on a system, only for test

CONF = cfg.CONF


class CommandSimulator(object):
    def __init__(self, pool_name):
        self._all_pools_name = {'acs5000_volpool_name': pool_name}
        self._pools_list = {
            'pool01': {
                'name': 'pool01',
                'capacity': '799090409472',
                'free_capacity': '795869184000',
                'used_capacity': '3221225472',
                'total_volumes': 0},
            'pool02': {
                'name': 'pool02',
                'capacity': '193273528320',
                'free_capacity': '190052302848',
                'used_capacity': '3221225472',
                'total_volumes': 0
            }}
        self._volumes_list = {}
        self._lun_maps_list = []
        self._snapshots_list = []
        self._controllers_list = [
            {'id': '0',
             'name': 'node1',
             'iscsi_name': 'iqn.2020-12.cn.com.toyou:'
                           'disk-array-000f12345:dev0.ctr1',
             'WWNN': '200008CA45D33768',
             'status': 'online'},
            {'id': '1',
             'name': 'node2',
             'iscsi_name': 'iqn.2020-04.cn.com.toyou:'
                           'disk-array-000f12345:dev0.ctr2',
             'WWNN': '200008CA45D33768',
             'status': 'online'}]
        self._ip_list = {
            '0': [{
                'ctrl_idx': 0,
                'id': 0,
                'name': 'lan1',
                'ip': '10.23.45.67',
                'mask': '255.255.255.0',
                'gw': ''
            }, {
                'ctrl_idx': 0,
                'id': 1,
                'name': 'lan2',
                'ip': '10.23.45.68',
                'mask': '255.255.255.0',
                'gw': ''
            }],
            '1': [{
                'ctrl_idx': 1,
                'id': 0,
                'name': 'lan1',
                'ip': '10.23.45.69',
                'mask': '255.255.255.0',
                'gw': ''
            }, {
                'ctrl_idx': 1,
                'id': 1,
                'name': 'lan2',
                'ip': '10.23.45.70',
                'mask': '255.255.255.0',
                'gw': ''
            }]
        }
        self._system_info = {'version': '3.1.2.345678',
                             'vendor': 'TOYOU',
                             'system_name': 'Disk-Array',
                             'system_id': 'TY123456789ABCDEF',
                             'code_level': '1',
                             'ip': '10.0.0.1'}

        self._error = {
            'success': (0, 'Success'),
            'unknown': (1, 'unknown error'),
            'pool_not_exist': (101, 'The pool does not exist '
                                    'on the system.'),
            'pool_exceeds_size': (102, 'The pool cannot provide '
                                       'more storage space'),
            'volume_not_exist': (303, 'The volume does not exist '
                                      'on the system.'),
            'source_volume_not_exist': (304, 'A clone relation needs '
                                             'a source volume.'),
            'target_volume_not_exist': (305, 'A clone relation needs '
                                             'a target volume.'),
            'source_size_larger_target': (306, 'The source volume '
                                               'must not be larger '
                                               'than the target volume'
                                               ' in a clone relation '),
            'volume_limit_pool': (307, 'A pool only supports 96 volumes'),
            'volume_limit_system': (308, 'A system only supports 96 volumes'),
            'volume_name_exist': (310, 'A volume with same name '
                                       'already exists on the system.'),
            'volume_extend_min': (321, 'A volume capacity shall not be'
                                       ' less than the current size'),
            'lun_not_exist': (401, 'The volume does not exist  '
                                   'on the system.'),
            'not_available_lun': (402, 'The system have no available lun.'),
            'snap_over_system': (503, 'The system snapshots maximum quantity '
                                      'has been reached.'),
            'snap_over_volume': (504, 'A volume snapshots maximum quantity '
                                      'has been reached.'),
            'snap_not_exist': (505, 'The snapshot does not exist '
                                    'on the system.')
        }
        self._command_function = {
            'sshGetSystem': 'get_system',
            'sshGetIpConnect': 'get_ip_connect',
            'sshGetPoolInfo': 'get_pool_info',
            'sshGetVolume': 'get_volume',
            'sshGetCtrInfo': 'ls_ctr_info',
            'sshCreateVolume': 'create_volume',
            'sshDeleteVolume': 'delete_volume',
            'sshCinderExtendVolume': 'extend_volume',
            'sshMkLocalClone': 'create_clone',
            'sshMkStartLocalClone': 'start_clone',
            'sshRemoveLocalClone': 'delete_clone',
            'sshMapVoltoHost': 'create_lun_map',
            'sshDeleteLunMap': 'delete_lun_map',
            'sshCreateSnapshot': 'create_snapshot',
            'sshDeleteSnapshot': 'delete_snapshot',
            'sshSetVolumeProperty': 'set_volume_property',
            'error_ssh': 'error_ssh'
        }
        self._volume_type = {
            '0': 'RAID Volume',
            '10': 'BACKUP'
        }

    @staticmethod
    def _json_return(rows=None, msg='', key=0):
        json_data = {'key': key,
                     'msg': msg,
                     'arr': rows}
        return (json.dumps(json_data), '')

    @staticmethod
    def _create_id(lists, key='id'):
        ids = []
        if isinstance(lists, list):
            for v in lists:
                ids.append(int(v[key]))
        elif isinstance(lists, dict):
            for v in lists.values():
                ids.append(int(v[key]))
        new_id = 'ffffffffff'
        while True:
            new_id = str(random.randint(1000000000, 9999999999))
            if new_id not in ids:
                break
        return new_id

    def _clone_thread(self, vol_name, setting=None):
        intval = 0.1
        loop_times = int(self._volumes_list[vol_name]['size_gb'])
        chunk = int(100 / loop_times)
        if setting:
            for k, value in setting.items():
                for v in value:
                    self._volumes_list[k][v[0]] = v[1]
                    time.sleep(v[2])

        self._volumes_list[vol_name]['status'] = 'Cloning'
        while loop_times > 0:
            # volumes may be deleted
            if vol_name in self._volumes_list:
                src_vol = self._volumes_list[vol_name]
            else:
                return
            if src_vol['clone'] not in self._volumes_list:
                self._volumes_list[vol_name]['status'] = 'Online'
                self._volumes_list[vol_name]['r'] = ''
                return
            progress = src_vol['r']
            if not progress:
                progress = 0
            src_vol['r'] = str(int(progress) + chunk)
            loop_times -= 1
            self._volumes_list[vol_name] = src_vol
            time.sleep(intval)
        self._volumes_list[vol_name]['status'] = 'Online'
        self._volumes_list[vol_name]['r'] = ''

    def execute_command(self, cmd_list, check_exit_code=True):
        command = cmd_list[2]
        if command in self._command_function:
            command = self._command_function[command]
        func = getattr(self, '_sim_' + command)
        kwargs = {}
        for i in range(3, len(cmd_list)):
            if cmd_list[i].startswith('--'):
                key = cmd_list[i][2:]
                value = ''
                if cmd_list[i + 1]:
                    value = cmd_list[i + 1]
                    i += 1
                if key in kwargs.keys():
                    if not isinstance(kwargs[key], list):
                        kwargs[key] = [kwargs[key]]

                    kwargs[key].append(value)
                else:
                    kwargs[key] = value
        try:
            out, err = func(**kwargs)
            return (out, err)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                if check_exit_code:
                    raise processutils.ProcessExecutionError(
                        exit_code=1,
                        stdout='out',
                        stderr=e,
                        cmd=' '.join(cmd_list))

    def _sim_get_system(self, **kwargs):
        return self._json_return(self._system_info)

    def _sim_get_ip_connect(self, **kwargs):
        return self._json_return(self._ip_list)

    def _sim_get_pool_info(self, **kwargs):
        pool_name = kwargs['poolName'].strip('\'\"')
        if pool_name in self._all_pools_name['acs5000_volpool_name']:
            vol_len = 0
            for vol in self._volumes_list.values():
                if vol['poolname'] == pool_name:
                    vol_len += 1
            if pool_name in self._pools_list:
                pool_data = self._pools_list[pool_name]
            else:
                pool_data = self._pools_list['pool01']
                pool_data['name'] = pool_name
            pool_data['total_volumes'] = str(vol_len)
            return self._json_return(pool_data)
        else:
            return self._json_return()

    def _sim_get_volume(self, **kwargs):
        rows = []
        if isinstance(kwargs['name'], list):
            volume_name = kwargs['name']
        else:
            volume_name = [kwargs['name']]
        for vol_name in volume_name:
            if vol_name in self._volumes_list.keys():
                rows.append(self._volumes_list[vol_name])

        return self._json_return(rows)

    def _sim_ls_ctr_info(self, **kwargs):
        return self._json_return(self._controllers_list)

    def _sim_create_volume(self, **kwargs):
        volume_name = kwargs['volumename']
        pool_name = kwargs['cinderPool']
        size = kwargs['volumesize']
        if volume_name in self._volumes_list:
            return self._json_return(
                msg=self._error['volume_name_exist'][1],
                key=self._error['volume_name_exist'][0])
        elif len(self._volumes_list) >= VOLUME_LIMIT_ON_SYSTEM:
            return self._json_return(
                msg=self._error['volume_limit_system'][1],
                key=self._error['volume_limit_system'][0])

        volume_count_on_pool = 0
        for v in self._volumes_list.values():
            if v['poolname'] == pool_name:
                volume_count_on_pool += 1
        if volume_count_on_pool >= VOLUME_LIMIT_ON_POOL:
            return self._json_return(
                msg=self._error['volume_limit_pool'][1],
                key=self._error['volume_limit_pool'][0])
        avail_size = (int(self._pools_list[pool_name]['free_capacity'])
                      / units.Gi)
        if int(size) > avail_size:
            return self._json_return(
                msg=self._error['pool_exceeds_size'][1],
                key=self._error['pool_exceeds_size'][0])
        volume_info = {}
        volume_info['id'] = self._create_id(self._volumes_list)
        volume_info['name'] = volume_name
        volume_info['size_gb'] = size
        volume_info['status'] = 'Online'
        volume_info['health'] = 'Optimal'
        volume_info['r'] = ''
        volume_info['poolname'] = pool_name
        volume_info['has_clone'] = 0
        volume_info['clone'] = 'N/A'
        volume_info['clone_snap'] = 'N/A'
        type = kwargs['type']
        if type not in ('0', '10'):
            type = '0'
        volume_info['type'] = self._volume_type[type]
        self._volumes_list[volume_info['name']] = volume_info
        return self._json_return()

    def _sim_delete_volume(self, **kwargs):
        vol_name = kwargs['cinderVolume']
        if vol_name in self._volumes_list:
            del self._volumes_list[vol_name]
        return self._json_return()

    def _sim_extend_volume(self, **kwargs):
        vol_name = kwargs['cinderVolume']
        size = int(kwargs['extendsize'])
        if vol_name not in self._volumes_list:
            return self._json_return(
                msg=self._error['volume_not_exist'][1],
                key=self._error['volume_not_exist'][0])
        volume = self._volumes_list[vol_name]
        curr_size = int(volume['size_gb'])
        pool = self._pools_list[volume['poolname']]
        avail_size = int(pool['free_capacity']) / units.Gi
        if curr_size > size:
            return self._json_return(
                msg=self._error['volume_extend_min'][1],
                key=self._error['volume_extend_min'][0])
        elif (size - curr_size) > avail_size:
            return self._json_return(
                msg=self._error['pool_exceeds_size'][1],
                key=self._error['pool_exceeds_size'][0])
        self._volumes_list[vol_name]['size_gb'] = str(size)
        return self._json_return()

    def _sim_create_clone(self, **kwargs):
        src_name = kwargs['cinderVolume']
        tgt_name = kwargs['cloneVolume']
        src_exist = False
        tgt_exist = False
        for vol in self._volumes_list.values():
            if (vol['name'] == src_name
                    and vol['type'] == self._volume_type['0']):
                src_exist = True
            elif (vol['name'] == tgt_name
                    and vol['type'] == self._volume_type['10']):
                tgt_exist = True
            if src_exist and tgt_exist:
                break
        if not src_exist:
            return self._json_return(
                msg=self._error['source_volume_not_exist'][1],
                key=self._error['source_volume_not_exist'][0])
        elif not tgt_exist:
            return self._json_return(
                msg=self._error['target_volume_not_exist'][1],
                key=self._error['target_volume_not_exist'][0])
        src_size = int(self._volumes_list[src_name]['size_gb'])
        tgt_size = int(self._volumes_list[tgt_name]['size_gb'])
        if src_size > tgt_size:
            return self._json_return(
                msg=self._error['source_size_larger_target'][1],
                key=self._error['source_size_larger_target'][0])
        tgt_volume = self._volumes_list[tgt_name]
        self._volumes_list[src_name]['has_clone'] = 1
        self._volumes_list[src_name]['clone'] = tgt_volume['name']
        return self._json_return()

    def _sim_start_clone(self, **kwargs):
        vol_name = kwargs['cinderVolume']
        snapshot = kwargs['snapshot']
        if len(snapshot) > 0:
            snap_found = False
            for snap in self._snapshots_list:
                if snap['name'] == snapshot:
                    snap_found = True
                    break
            if not snap_found:
                return self._json_return(
                    msg=self._error['snap_not_exist'][1],
                    key=self._error['snap_not_exist'][0])
        else:
            snapshot = ('clone-' + str(random.randint(100, 999)))
            tmp_snap = {'volume': vol_name,
                        'snapshot': snapshot}
            self._sim_create_snapshot(**tmp_snap)
        self._volumes_list[vol_name]['status'] = 'Queued'
        self._volumes_list[vol_name]['clone_snap'] = snapshot
        greenthread.spawn_n(self._clone_thread, vol_name)
        return self._json_return()

    def _sim_delete_clone(self, **kwargs):
        vol_name = kwargs['name']
        snapshot = kwargs['snapshot']
        if vol_name not in self._volumes_list:
            return self._json_return(
                msg=self._error['volume_not_exist'][1],
                key=self._error['volume_not_exist'][0])
        self._volumes_list[vol_name]['has_clone'] = 0
        clone_volume = self._volumes_list[vol_name]['clone']
        self._volumes_list[vol_name]['clone'] = 'N/A'
        clone_snap = self._volumes_list[vol_name]['clone_snap']
        self._volumes_list[vol_name]['clone_snap'] = 'N/A'
        self._volumes_list[clone_volume]['type'] = self._volume_type['0']
        if len(snapshot) == 0:
            for snap in self._snapshots_list:
                if clone_snap == snap['name']:
                    self._snapshots_list.remove(snap)
                    break
        return self._json_return()

    def _sim_create_lun_map(self, **kwargs):
        volume_name = kwargs['cinderVolume']
        protocol = kwargs['protocol']
        hosts = kwargs['host']
        if volume_name not in self._volumes_list:
            return self._json_return(
                msg=self._error['volume_not_exist'][1],
                key=self._error['volume_not_exist'][0])
        if isinstance(hosts, str):
            hosts = [hosts]
        volume = self._volumes_list[volume_name]
        available_luns = LUN_NUMS_AVAILABLE
        existed_lun = -1
        for lun_row in self._lun_maps_list:
            if lun_row['vd_id'] == volume['id']:
                if lun_row['host'] in hosts:
                    existed_lun = lun_row['lun']
                    hosts = [h for h in hosts if h != lun_row['host']]
            else:
                if lun_row['protocol'] == protocol:
                    available_luns = [lun for lun in available_luns
                                      if lun != lun_row['lun']]
        if hosts and existed_lun > -1:
            return self._json_return({'info': existed_lun})
        lun_info = {}
        lun_info['vd_id'] = volume['id']
        lun_info['vd_name'] = volume['name']
        lun_info['protocol'] = protocol
        if existed_lun > -1:
            lun_info['lun'] = existed_lun
        elif available_luns:
            lun_info['lun'] = available_luns[0]
        else:
            return self._json_return(
                msg=self._error['not_available_lun'][1],
                key=self._error['not_available_lun'][0])
        for host in hosts:
            lun_info['id'] = self._create_id(self._lun_maps_list)
            lun_info['host'] = host
            self._lun_maps_list.append(copy.deepcopy(lun_info))
        ret = {'lun': [],
               'iscsi_name': [],
               'portal': []}
        for ctr, ips in self._ip_list.items():
            for ip in ips:
                ret['lun'].append(lun_info['lun'])
                ret['portal'].append('%s:3260' % ip['ip'])
                for control in self._controllers_list:
                    if ctr == control['id']:
                        ret['iscsi_name'].append(control['iscsi_name'])
                        break
        return self._json_return(ret)

    def _sim_delete_lun_map(self, **kwargs):
        map_exist = False
        volume_name = kwargs['cinderVolume']
        protocol = kwargs['protocol']
        hosts = kwargs['cinderHost']
        if isinstance(hosts, str):
            hosts = [hosts]
        if volume_name not in self._volumes_list:
            return self._json_return(
                msg=self._error['volume_not_exist'][1],
                key=self._error['volume_not_exist'][0])
        volume = self._volumes_list[volume_name]
        lun_maps_list = self._lun_maps_list
        self._lun_maps_list = []
        for row in lun_maps_list:
            if (row['vd_id'] == volume['id']
                    and row['protocol'] == protocol
                    and row['host'] in hosts):
                map_exist = True
            else:
                map_exist = False
                self._lun_maps_list.append(row)
        if not map_exist:
            return self._json_return(
                msg=self._error['lun_not_exist'][1],
                key=self._error['lun_not_exist'][0])
        else:
            return self._json_return()

    def _sim_create_snapshot(self, **kwargs):
        volume_name = kwargs['volume']
        snapshot_name = kwargs['snapshot']
        if volume_name not in self._volumes_list:
            return self._json_return(
                msg=self._error['volume_not_exist'][1],
                key=self._error['volume_not_exist'][0])
        if len(self._snapshots_list) >= SNAPSHOTS_ON_SYSTEM:
            return self._json_return(
                msg=self._error['snap_over_system'][1],
                key=self._error['snap_over_system'][0])
        tag = -1
        volume_snap_count = 0
        for snap in self._snapshots_list:
            if snap['vd_name'] == volume_name:
                volume_snap_count += 1
                if int(snap['tag']) > tag:
                    tag = int(snap['tag'])
        if volume_snap_count >= SNAPSHOTS_A_VOLUME:
            return self._json_return(
                msg=self._error['snap_over_volume'][1],
                key=self._error['snap_over_volume'][0])
        volume = self._volumes_list[volume_name]
        snapshot = {}
        snapshot['id'] = self._create_id(self._snapshots_list)
        snapshot['name'] = snapshot_name
        snapshot['vd_id'] = volume['id']
        snapshot['vd_name'] = volume['name']
        snapshot['tag'] = tag + 1
        snapshot['create_time'] = ''
        self._snapshots_list.append(snapshot)
        return self._json_return()

    def _sim_delete_snapshot(self, **kwargs):
        volume_name = kwargs['volume']
        snapshot_name = kwargs['snapshot']
        if volume_name not in self._volumes_list:
            return self._json_return(
                msg=self._error['volume_not_exist'][1],
                key=self._error['volume_not_exist'][0])
        snap_exist = False
        for snap in self._snapshots_list:
            if (snap['vd_name'] == volume_name
                    and snap['name'] == snapshot_name):
                snap_exist = True
                self._snapshots_list.remove(snap)
                break
        if not snap_exist:
            return self._json_return(
                msg=self._error['snap_not_exist'][1],
                key=self._error['snap_not_exist'][0])
        return self._json_return()

    def _sim_set_volume_property(self, **kwargs):
        volume_name = kwargs['volume']
        kwargs.pop('volume')
        if len(kwargs) == 0:
            raise exception.InvalidInput(
                reason=self._error['unknown'][1])
        new_name = volume_name
        if 'new_name' in kwargs:
            new_name = kwargs['new_name']
            kwargs.pop('new_name')
        if volume_name not in self._volumes_list:
            return self._json_return(
                msg=self._error['volume_not_exist'][1],
                key=self._error['volume_not_exist'][0])
        volume = self._volumes_list[volume_name]
        volume['name'] = new_name
        for k, v in kwargs.items():
            if k in volume:
                volume[k] = v
            else:
                return ('', self._error['unknown'][1])

        if volume_name != new_name:
            del self._volumes_list[volume_name]
            self._volumes_list[new_name] = volume
        else:
            self._volumes_list[volume_name] = volume
        return self._json_return()

    def _sim_error_ssh(self, **kwargs):
        error = kwargs['error']
        if error == 'json_error':
            return ('This text is used for json errors.', '')
        elif error == 'dict_error':
            return (json.dumps('This text is used for dict errors.'), '')
        elif error == 'keys_error':
            keys = {'msg': 'This text is used for keys errors'}
            return (json.dumps(keys), '')
        elif error == 'key_false':
            keys = {'msg': 'This text is used for key non-0 error',
                    'key': 1,
                    'arr': {}}
            return (json.dumps(keys), '')


class Acs5000ISCSIFakeDriver(acs5000_iscsi.Acs5000ISCSIDriver):
    def __init__(self, *args, **kwargs):
        super(Acs5000ISCSIFakeDriver, self).__init__(*args, **kwargs)

    def set_fake_storage(self, fake):
        self.fake_storage = fake

    def _run_ssh(self, cmd_list, check_exit_code=True):
        cinder_utils.check_ssh_injection(cmd_list)
        ret = self.fake_storage.execute_command(cmd_list, check_exit_code)

        return ret


class Acs5000ISCSIDriverTestCase(test.TestCase):
    @mock.patch.object(time, 'sleep')
    def setUp(self, mock_sleep):
        super(Acs5000ISCSIDriverTestCase, self).setUp()
        self.configuration = mock.Mock(conf.Configuration)
        self.configuration.san_is_local = False
        self.configuration.san_ip = '23.44.56.78'
        self.configuration.san_login = 'cliuser'
        self.configuration.san_password = 'clipassword'
        self.configuration.acs5000_volpool_name = ['pool01']
        self.iscsi_driver = Acs5000ISCSIFakeDriver(
            configuration=self.configuration)
        initiator = 'test.iqn.%s' % str(random.randint(10000, 99999))
        self._connector = {'ip': '1.234.56.78',
                           'host': 'stack',
                           'wwpns': [],
                           'initiator': initiator}
        self.sim = CommandSimulator(POOLS_NAME)
        self.iscsi_driver.set_fake_storage(self.sim)
        self.ctxt = context.get_admin_context()

        self.db = cinder.db
        self.iscsi_driver.db = self.db
        self.iscsi_driver.get_driver_options()
        self.iscsi_driver.do_setup(None)
        self.iscsi_driver.check_for_setup_error()

    def _create_volume(self, **kwargs):
        prop = {'host': 'stack@ty1#%s' % POOLS_NAME[0],
                'size': 1,
                'volume_type_id': self.vt['id']}
        for p in prop.keys():
            if p not in kwargs:
                kwargs[p] = prop[p]
        vol = testutils.create_volume(self.ctxt, **kwargs)
        self.iscsi_driver.create_volume(vol)
        return vol

    def _delete_volume(self, volume):
        self.iscsi_driver.delete_volume(volume)
        self.db.volume_destroy(self.ctxt, volume['id'])

    def _assert_lun_exists(self, vol_id, exists):
        lun_maps = self.sim._lun_maps_list
        is_lun_defined = False
        luns = []
        volume_name = VOLUME_PRE + vol_id[-12:]
        for lun in lun_maps:
            if volume_name == lun['vd_name']:
                luns.append(lun)
        if len(luns):
            is_lun_defined = True
        self.assertEqual(exists, is_lun_defined)
        return luns

    def test_validate_connector(self):
        conn_neither = {'host': 'host'}
        conn_iscsi = {'host': 'host', 'initiator': 'iqn.123'}
        conn_fc = {'host': 'host', 'wwpns': 'fff123'}
        conn_both = {'host': 'host', 'initiator': 'iqn.123', 'wwpns': 'fff123'}

        self.iscsi_driver._state['enabled_protocols'] = set(['iSCSI'])
        self.iscsi_driver.validate_connector(conn_iscsi)
        self.iscsi_driver.validate_connector(conn_both)
        self.assertRaises(exception.InvalidConnectorException,
                          self.iscsi_driver.validate_connector, conn_fc)
        self.assertRaises(exception.InvalidConnectorException,
                          self.iscsi_driver.validate_connector, conn_neither)

    def test_initialize_connection(self):
        volume = self._create_volume()
        result = self.iscsi_driver.initialize_connection(volume,
                                                         self._connector)
        ip_connect = self.sim._ip_list
        ip_count = 0
        for ips in ip_connect.values():
            ip_count += len(ips)
        self.assertEqual('iscsi', result['driver_volume_type'])
        self.assertEqual(ip_count,
                         len(result['data']['target_iqns']))
        self.assertEqual(ip_count,
                         len(result['data']['target_portals']))
        self.assertEqual(volume['id'], result['data']['volume_id'])
        self.assertEqual(ip_count,
                         len(result['data']['target_portals']))
        self._delete_volume(volume)

    def test_initialize_connection_not_found(self):
        prop = {'host': 'stack@ty1#%s' % POOLS_NAME[0],
                'size': 1,
                'volume_type_id': self.vt['id']}
        vol = testutils.create_volume(self.ctxt, **prop)
        self.assertRaises(exception.VolumeNotFound,
                          self.iscsi_driver.initialize_connection,
                          vol, self._connector)
        self.db.volume_destroy(self.ctxt, vol['id'])

    def test_initialize_connection_failure(self):
        volume_list = []
        for i in LUN_NUMS_AVAILABLE:
            vol = self._create_volume()
            self.iscsi_driver.initialize_connection(
                vol, self._connector)
            volume_list.append(vol)

        vol = self._create_volume()
        self.assertRaises(exception.ISCSITargetAttachFailed,
                          self.iscsi_driver.initialize_connection,
                          vol, self._connector)
        self._delete_volume(vol)
        for v in volume_list:
            self.iscsi_driver.terminate_connection(
                v, self._connector)
            self._delete_volume(v)

    def test_initialize_connection_multi_host(self):
        connector = self._connector
        initiator1 = ('test.iqn.%s'
                      % str(random.randint(10000, 99999)))
        initiator2 = ('test.iqn.%s'
                      % str(random.randint(10000, 99999)))
        connector['initiator'] = [initiator1, initiator2]
        volume = self._create_volume()
        self.iscsi_driver.initialize_connection(
            volume, connector)
        lun_maps = self._assert_lun_exists(volume['id'], True)
        hosts = []
        for lun in lun_maps:
            hosts.append(lun['host'])
        self.assertIn(initiator1, hosts)
        self.assertIn(initiator2, hosts)
        self.iscsi_driver.terminate_connection(
            volume, connector)
        self._assert_lun_exists(volume['id'], False)
        self._delete_volume(volume)

    def test_terminate_connection(self):
        volume = self._create_volume()
        self.iscsi_driver.initialize_connection(volume,
                                                self._connector)
        self.iscsi_driver.terminate_connection(volume,
                                               self._connector)
        self._assert_lun_exists(volume['id'], False)
        self._delete_volume(volume)


class Acs5000CommonDriverTestCase(test.TestCase):
    @mock.patch.object(time, 'sleep')
    def setUp(self, mock_sleep):
        super(Acs5000CommonDriverTestCase, self).setUp()

        self.configuration = mock.Mock(conf.Configuration)
        self.configuration.san_is_local = False
        self.configuration.san_ip = '23.44.56.78'
        self.configuration.san_login = 'cliuser'
        self.configuration.san_password = 'clipassword'
        self.configuration.acs5000_volpool_name = POOLS_NAME
        self.configuration.acs5000_copy_interval = 0.01
        self.configuration.reserved_percentage = 0
        self._driver = Acs5000ISCSIFakeDriver(
            configuration=self.configuration)
        options = acs5000_iscsi.Acs5000ISCSIDriver.get_driver_options()
        config = conf.Configuration(options, conf.SHARED_CONF_GROUP)
        self.override_config('san_ip', '23.44.56.78', conf.SHARED_CONF_GROUP)
        self.override_config('san_login', 'cliuser', conf.SHARED_CONF_GROUP)
        self.override_config('san_password', 'clipassword',
                             conf.SHARED_CONF_GROUP)
        self.override_config('acs5000_volpool_name', POOLS_NAME,
                             conf.SHARED_CONF_GROUP)
        self._iscsi_driver = acs5000_iscsi.Acs5000ISCSIDriver(
            configuration=config)
        initiator = 'test.iqn.%s' % str(random.randint(10000, 99999))
        self._connector = {'ip': '1.234.56.78',
                           'host': 'stack',
                           'wwpns': [],
                           'initiator': initiator}
        self.sim = CommandSimulator(POOLS_NAME)
        self._driver.set_fake_storage(self.sim)
        self.ctxt = context.get_admin_context()

        self.db = cinder.db
        self._driver.db = self.db
        self._driver.do_setup(None)
        self._driver.check_for_setup_error()

    def _assert_vol_exists(self, name, exists):
        volume = self._driver._cmd.get_volume(VOLUME_PRE + name[-12:])
        is_vol_defined = False
        if volume:
            is_vol_defined = True
        self.assertEqual(exists, is_vol_defined)
        return volume

    def _assert_snap_exists(self, name, exists):
        snap_name = VOLUME_PRE + name[-12:]
        snapshot_list = self.sim._snapshots_list
        is_snap_defined = False
        snapshot = {}
        for snap in snapshot_list:
            if snap['name'] == snap_name:
                is_snap_defined = True
                snapshot = snap
                break
        self.assertEqual(exists, is_snap_defined)
        return snapshot

    def _create_volume(self, **kwargs):
        prop = {'host': 'stack@ty1#%s' % POOLS_NAME[0],
                'size': 1,
                'volume_type_id': self.vt['id']}
        driver = True
        if 'driver' in kwargs:
            if not kwargs['driver']:
                driver = False
            kwargs.pop('driver')
        for p in prop.keys():
            if p not in kwargs:
                kwargs[p] = prop[p]
        vol = testutils.create_volume(self.ctxt, **kwargs)
        if driver:
            self._driver.create_volume(vol)
        return vol

    def _delete_volume(self, volume, driver=True):
        if driver:
            self._driver.delete_volume(volume)
        self.db.volume_destroy(self.ctxt, volume['id'])

    def _create_snapshot(self, vol_id, driver=True):
        snap = testutils.create_snapshot(self.ctxt, vol_id)
        if driver:
            self._driver.create_snapshot(snap)
        return snap

    def _delete_snapshot(self, snap, driver=True):
        if driver:
            self._driver.delete_snapshot(snap)
        self.db.snapshot_destroy(self.ctxt, snap['id'])

    def test_run_ssh_failure(self):
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._build_pool_stats,
                          'error_pool')
        ssh_cmd = ['cinder', 'storage', 'error_ssh', '--error', 'json_error']
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._cmd.run_ssh_info, ssh_cmd)
        ssh_cmd = ['cinder', 'storage', 'error_ssh', '--error', 'dict_error']
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._cmd.run_ssh_info, ssh_cmd)
        ssh_cmd = ['cinder', 'storage', 'error_ssh', '--error', 'keys_error']
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._cmd.run_ssh_info, ssh_cmd)
        ssh_cmd = ['cinder', 'storage', 'error_ssh', '--error', 'key_false']
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._cmd.run_ssh_info, ssh_cmd)

    @mock.patch.object(ssh_utils, 'SSHPool')
    @mock.patch.object(processutils, 'ssh_execute')
    def test_run_ssh_with_ip(self, mock_ssh_execute, mock_ssh_pool):
        ssh_cmd = ['cinder', 'storage', 'run_ssh']
        self._iscsi_driver._run_ssh(ssh_cmd)
        mock_ssh_pool.assert_called_once_with(
            self._iscsi_driver.configuration.san_ip,
            self._iscsi_driver.configuration.san_ssh_port,
            self._iscsi_driver.configuration.ssh_conn_timeout,
            self._iscsi_driver.configuration.san_login,
            password=self._iscsi_driver.configuration.san_password,
            min_size=self._iscsi_driver.configuration.ssh_min_pool_conn,
            max_size=self._iscsi_driver.configuration.ssh_max_pool_conn)

        mock_ssh_pool.side_effect = [paramiko.SSHException, mock.MagicMock()]
        self._iscsi_driver._run_ssh(ssh_cmd)
        mock_ssh_pool.assert_called_once_with(
            self._iscsi_driver.configuration.san_ip,
            self._iscsi_driver.configuration.san_ssh_port,
            self._iscsi_driver.configuration.ssh_conn_timeout,
            self._iscsi_driver.configuration.san_login,
            password=self._iscsi_driver.configuration.san_password,
            min_size=self._iscsi_driver.configuration.ssh_min_pool_conn,
            max_size=self._iscsi_driver.configuration.ssh_max_pool_conn)

    @mock.patch.object(ssh_utils, 'SSHPool')
    @mock.patch.object(processutils, 'ssh_execute')
    def test_run_ssh_with_exception(self, mock_ssh_execute, mock_ssh_pool):
        mock_ssh_execute.side_effect = [processutils.ProcessExecutionError,
                                        mock.MagicMock()]
        self.override_config('acs5000_volpool_name', None,
                             self._iscsi_driver.configuration.config_group)
        ssh_cmd = ['cinder', 'storage', 'run_ssh']
        self.assertRaises(processutils.ProcessExecutionError,
                          self._iscsi_driver._run_ssh, ssh_cmd)

    def test_do_setup(self):
        system_info = self.sim._system_info
        self.assertEqual(system_info['vendor'], self._driver._state['vendor'])
        self.assertIn('iSCSI', self._driver._state['enabled_protocols'])
        self.assertEqual(2, len(self._driver._state['storage_nodes']))

    def test_do_setup_no_pools(self):
        self._driver.pools = ['pool_error']
        self.assertRaises(exception.InvalidInput,
                          self._driver.do_setup, None)

    def test_create_volume(self):
        vol = self._create_volume()
        self._assert_vol_exists(vol['id'], True)
        self._delete_volume(vol)

    def test_create_volume_same_name(self):
        vol = self._create_volume()
        self._assert_vol_exists(vol['id'], True)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver.create_volume, vol)
        self._delete_volume(vol)

    def test_create_volume_size_exceeds_limit(self):
        prop = {
            'host': 'stack@ty2#%s' % POOLS_NAME[1],
            'size': 200,
            'driver': False
        }
        self._driver.get_volume_stats()
        vol = self._create_volume(**prop)
        self._assert_vol_exists(vol['id'], False)
        self.assertRaises(exception.VolumeSizeExceedsLimit,
                          self._driver.create_volume,
                          vol)
        self._delete_volume(vol, False)

    def test_create_volume_number_exceeds_pool_limit(self):
        volume_list = []
        for i in range(VOLUME_LIMIT_ON_POOL):
            vol = self._create_volume()
            self._assert_vol_exists(vol['id'], True)
            volume_list.append(vol)
        vol = self._create_volume(driver=False)
        self.assertRaises(exception.VolumeLimitExceeded,
                          self._driver.create_volume,
                          vol)
        self._delete_volume(vol, False)
        for v in volume_list:
            self._delete_volume(v)

    def test_create_volume_number_exceeds_system_limit(self):
        volume_list = []
        volume_count_on_pool = int(VOLUME_LIMIT_ON_SYSTEM
                                   / len(POOLS_NAME))
        for i in range(volume_count_on_pool):
            for x in range(len(POOLS_NAME)):
                vol = self._create_volume(
                    host='stack@ty1#%s' % POOLS_NAME[x])
                self._assert_vol_exists(vol['id'], True)
                volume_list.append(vol)
        vol = self._create_volume(driver=False)
        self.assertRaises(exception.VolumeLimitExceeded,
                          self._driver.create_volume,
                          vol)
        self._delete_volume(vol, False)
        for v in volume_list:
            self._delete_volume(v)

    def test_delete_volume(self):
        vol = self._create_volume()
        self._assert_vol_exists(vol['id'], True)
        self._delete_volume(vol)
        self._assert_vol_exists(vol['id'], False)

    def test_create_snapshot(self):
        vol = self._create_volume()
        self._assert_vol_exists(vol['id'], True)
        snap = self._create_snapshot(vol['id'])
        self._assert_snap_exists(snap['id'], True)
        self._delete_snapshot(snap)
        self._delete_volume(vol)

    def test_create_snapshot_exceed_limit(self):
        vol = self._create_volume()
        self._assert_vol_exists(vol['id'], True)
        snapshot_list = []
        for i in range(SNAPSHOTS_A_VOLUME):
            snap = self._create_snapshot(vol['id'])
            self._assert_snap_exists(snap['id'], True)
            snapshot_list.append(snap)
        snap = self._create_snapshot(vol['id'], False)
        self.assertRaises(exception.SnapshotLimitExceeded,
                          self._driver.create_snapshot, snap)
        self._delete_snapshot(snap, False)
        vol_list = [vol]
        snap_count = SNAPSHOTS_A_VOLUME
        while snap_count < SNAPSHOTS_ON_SYSTEM:
            vol = self._create_volume()
            vol_list.append(vol)
            for x in range(SNAPSHOTS_A_VOLUME):
                snap = self._create_snapshot(vol['id'])
                self._assert_snap_exists(snap['id'], True)
                snapshot_list.append(snap)
                snap_count += 1
                if snap_count >= SNAPSHOTS_ON_SYSTEM:
                    break

        vol = self._create_volume()
        vol_list.append(vol)
        snap = self._create_snapshot(vol['id'], False)
        self.assertRaises(exception.SnapshotLimitExceeded,
                          self._driver.create_snapshot, snap)
        for sp in snapshot_list:
            self._delete_snapshot(sp)
        for vol in vol_list:
            self._delete_volume(vol)

    def test_delete_snapshot(self):
        vol = self._create_volume()
        self._assert_vol_exists(vol['id'], True)
        snap = self._create_snapshot(vol['id'])
        self._assert_snap_exists(snap['id'], True)
        self._delete_snapshot(snap)
        self._assert_snap_exists(snap['id'], False)
        self._delete_volume(vol)

    def test_delete_snapshot_not_found(self):
        vol = self._create_volume()
        self._assert_vol_exists(vol['id'], True)
        snap = self._create_snapshot(vol['id'], False)
        self._assert_snap_exists(snap['id'], False)
        self.assertRaises(exception.SnapshotNotFound,
                          self._driver.delete_snapshot,
                          snap)
        self._delete_snapshot(snap, False)
        self._delete_volume(vol)

    def test_create_volume_from_snapshot(self):
        prop = {'size': 2}
        vol = self._create_volume(**prop)
        self._assert_vol_exists(vol['id'], True)
        snap = self._create_snapshot(vol['id'])
        self._assert_snap_exists(snap['id'], True)
        prop['driver'] = False
        new_vol = self._create_volume(**prop)
        self._driver.create_volume_from_snapshot(new_vol, snap)
        new_volume = self._assert_vol_exists(new_vol['id'], True)
        self.assertEqual(1, len(new_volume))
        self.assertEqual('2', new_volume[0]['size_gb'])
        self.assertEqual('RAID Volume', new_volume[0]['type'])
        self._delete_volume(new_vol)
        self._delete_snapshot(snap)
        self._delete_volume(vol)

    def test_create_volume_from_snapshot_not_found(self):
        vol = self._create_volume()
        self._assert_vol_exists(vol['id'], True)
        snap = self._create_snapshot(vol['id'], False)
        self._assert_snap_exists(snap['id'], False)
        new_vol = self._create_volume(driver=False)
        self._assert_vol_exists(new_vol['id'], False)
        self.assertRaises(exception.SnapshotNotFound,
                          self._driver.create_volume_from_snapshot,
                          new_vol, snap)
        self._delete_volume(new_vol, False)
        self._delete_snapshot(snap, False)
        self._delete_volume(vol)

    def test_create_snapshot_volume_not_found(self):
        vol = self._create_volume(driver=False)
        self._assert_vol_exists(vol['id'], False)
        self.assertRaises(exception.VolumeNotFound,
                          self._create_snapshot, vol['id'])
        self._delete_volume(vol, driver=False)

    def test_create_cloned_volume(self):
        src_volume = self._create_volume()
        self._assert_vol_exists(src_volume['id'], True)
        tgt_volume = self._create_volume(driver=False)
        self._driver.create_cloned_volume(tgt_volume, src_volume)
        volume = self._assert_vol_exists(tgt_volume['id'], True)
        self.assertEqual(1, len(volume))
        self.assertEqual('RAID Volume', volume[0]['type'])
        self._delete_volume(src_volume)
        self._delete_volume(tgt_volume)

    def test_create_cloned_volume_with_size(self):
        prop = {'size': 2}
        src_volume = self._create_volume(**prop)
        volume = self._assert_vol_exists(src_volume['id'], True)
        prop['driver'] = False
        tgt_volume = self._create_volume(**prop)
        self._driver.create_cloned_volume(tgt_volume, src_volume)
        clone_volume = self._assert_vol_exists(tgt_volume['id'], True)
        self.assertEqual(1, len(volume))
        self.assertEqual(1, len(clone_volume))
        self.assertEqual('RAID Volume', volume[0]['type'])
        self.assertEqual('RAID Volume', clone_volume[0]['type'])
        self.assertEqual('2', volume[0]['size_gb'])
        self.assertEqual('2', clone_volume[0]['size_gb'])
        self._delete_volume(src_volume)
        self._delete_volume(tgt_volume)

    def test_create_cloned_volume_size_failure(self):
        prop = {'size': 10}
        src_volume = self._create_volume(**prop)
        self._assert_vol_exists(src_volume['id'], True)
        prop = {'size': 5, 'driver': False}
        tgt_volume = self._create_volume(**prop)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver.create_cloned_volume,
                          tgt_volume, src_volume)
        self._assert_vol_exists(tgt_volume['id'], False)
        self._delete_volume(src_volume)
        self._delete_volume(tgt_volume, False)

    def test_create_cloned_volume_failure(self):
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._local_clone_copy,
                          None, None)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._local_clone_copy,
                          'src_test', 'tgt_test')
        src_volume = self._create_volume()
        src_name = VOLUME_PRE + src_volume['id'][-12:]
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._local_clone_copy,
                          src_name, 'tgt_test')
        self._delete_volume(src_volume)

    def test_wait_volume_copy(self):
        src_volume = self._create_volume(size=2)
        src_info = self._assert_vol_exists(src_volume['id'], True)[0]
        tgt_volume = self._create_volume(size=2)
        tgt_info = self._assert_vol_exists(tgt_volume['id'], True)[0]
        self._driver._cmd.set_volume_property(
            src_info['name'], {'status': 'Queued',
                               'clone_snap': tgt_info['name']})
        self._driver._cmd.set_volume_property(tgt_info['name'],
                                              {'type': 'BACKUP'})
        src_name = VOLUME_PRE + src_volume['id'][-12:]
        tgt_name = VOLUME_PRE + tgt_volume['id'][-12:]
        self._driver._cmd.create_clone(src_name, tgt_name)
        tgt_set = {
            tgt_name: [('status', 'Erasing', 0.2)],
            src_name: [('status', 'Erasing', 0.2)],
        }
        greenthread.spawn_n(self.sim._clone_thread,
                            src_name, tgt_set)
        ret = self._driver._wait_volume_copy(src_name, tgt_name,
                                             'test_func', 'test_action')
        self.assertTrue(ret)
        self._driver._cmd.set_volume_property(
            src_info['name'], {'status': 'error',
                               'clone_snap': tgt_info['name']})
        ret = self._driver._wait_volume_copy(src_name, tgt_name,
                                             'test_func', 'test_action')
        self.assertFalse(ret)
        self._driver._cmd.set_volume_property(
            src_info['name'], {'status': 'Online',
                               'clone_snap': tgt_info['name']})
        self._delete_volume(tgt_volume)
        self._assert_vol_exists(tgt_volume['id'], False)
        ret = self._driver._wait_volume_copy(src_name, tgt_name,
                                             'test_func', 'test_action')
        self.assertFalse(ret)
        self._driver._cmd.set_volume_property(src_info['name'],
                                              {'type': 'BACKUP'})
        ret = self._driver._wait_volume_copy(tgt_name, 'backup_test',
                                             'test_func', 'test_action')
        self.assertFalse(ret)
        self._delete_volume(src_volume)

    def test_extend_volume(self):
        volume = self._create_volume(size=10)
        vol_info = self._assert_vol_exists(volume['id'], True)
        self.assertEqual('10', vol_info[0]['size_gb'])
        self._driver.extend_volume(volume, '100')
        extend_vol = self._assert_vol_exists(volume['id'], True)
        self.assertEqual('100', extend_vol[0]['size_gb'])
        self._delete_volume(volume)

    def test_extend_volume_not_found(self):
        volume = self._create_volume(driver=False)
        self.assertRaises(exception.VolumeNotFound,
                          self._driver.extend_volume,
                          volume, 10)
        self._delete_volume(volume, False)

    def test_extend_volume_size_less(self):
        volume = self._create_volume(size=100)
        vol_info = self._assert_vol_exists(volume['id'], True)
        self.assertEqual('100', vol_info[0]['size_gb'])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver.extend_volume,
                          volume, '10')
        self._delete_volume(volume)

    def test_extend_volume_size_exceeds_limit(self):
        host = 'stack@ty2#%s' % POOLS_NAME[1]
        self._driver.get_volume_stats()
        volume = self._create_volume(size=10, host=host)
        vol_info = self._assert_vol_exists(volume['id'], True)
        self.assertEqual('10', vol_info[0]['size_gb'])
        self.assertEqual(POOLS_NAME[1], vol_info[0]['poolname'])
        self.assertRaises(exception.VolumeSizeExceedsLimit,
                          self._driver.extend_volume,
                          volume, '200')
        self._delete_volume(volume)

    def test_migrate_volume_same_pool(self):
        host = 'stack@ty1#%s' % POOLS_NAME[0]
        volume = self._create_volume(host=host)
        target_host = {
            'host': 'stack_new@ty1#%s' % POOLS_NAME[0],
            'capabilities': {
                'system_id': self.sim._system_info['system_id'],
                'pool_name': POOLS_NAME[0]
            }
        }
        ret = self._driver.migrate_volume(self.ctxt, volume, target_host)
        self.assertEqual((True, None), ret)

    def test_migrate_volume_different_system(self):
        host = 'stack@ty1#%s' % POOLS_NAME[0]
        volume = self._create_volume(host=host)
        target_host = {
            'host': 'stack_new@ty1#%s' % POOLS_NAME[0],
            'capabilities': {
                'system_id': 'test_system_id',
                'pool_name': POOLS_NAME[0]
            }
        }
        ret = self._driver.migrate_volume(self.ctxt, volume, target_host)
        self.assertEqual((False, None), ret)
        target_host = {
            'host': 'stack_new@ty1#%s' % POOLS_NAME[0],
            'capabilities': {
                'pool_name': POOLS_NAME[0]
            }
        }
        ret = self._driver.migrate_volume(self.ctxt, volume, target_host)
        self.assertEqual((False, None), ret)

    def test_migrate_volume_same_system_different_pool(self):
        host = 'stack@ty1#%s' % POOLS_NAME[0]
        volume = self._create_volume(host=host, size=2)
        target_host = {
            'host': 'stack_new@ty1#%s' % POOLS_NAME[1],
            'capabilities': {
                'system_id': self.sim._system_info['system_id'],
                'pool_name': POOLS_NAME[1]
            }
        }
        ret = self._driver.migrate_volume(self.ctxt, volume, target_host)
        self.assertEqual((True, None), ret)
        vol_info = self._assert_vol_exists(volume['id'], True)
        self.assertEqual(POOLS_NAME[1], vol_info[0]['poolname'])
        self.assertEqual('2', vol_info[0]['size_gb'])

    def test_get_volume_stats(self):
        self.assertEqual({}, self._driver._stats)
        self._driver.get_volume_stats()
        stats = self._driver._stats
        system_info = self.sim._system_info
        self.assertEqual(system_info['vendor'], stats['vendor_name'])

    def test_get_volume_none(self):
        ret = self._driver._cmd.get_volume('')
        self.assertEqual([], ret)
        ret = self._driver._cmd.get_volume('test_volume')
        self.assertEqual([], ret)

    def test_check_for_setup_error_failure(self):
        self._driver._state['system_name'] = None
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver.check_for_setup_error)
        self._driver.do_setup(None)
        self._driver._state['system_id'] = None
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver.check_for_setup_error)
        self._driver.do_setup(None)
        self._driver._state['storage_nodes'] = []
        self.assertRaises(exception.VolumeDriverException,
                          self._driver.check_for_setup_error)
        self._driver.do_setup(None)
        self._driver._state['enabled_protocols'] = set()
        self.assertRaises(exception.InvalidInput,
                          self._driver.check_for_setup_error)
        self._driver.do_setup(None)
        self._driver.configuration.san_password = None
        self.assertRaises(exception.InvalidInput,
                          self._driver.check_for_setup_error)
        self._driver.do_setup(None)

    def test_build_pool_stats_no_pool(self):
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._build_pool_stats,
                          'pool_test')

    def test_set_volume_property_failure(self):
        volume = self._create_volume()
        self._assert_vol_exists(volume['id'], True)
        volume_name = VOLUME_PRE + volume['id'][-12:]
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._cmd.set_volume_property,
                          volume_name, {'error_key': 'error'})
        self.assertRaises(exception.VolumeBackendAPIException,
                          self._driver._cmd.set_volume_property,
                          volume_name, {})
        self._delete_volume(volume)



# Copyright 2018 Red Hat, Inc
# Copyright 2017 Rackspace Australia
# Copyright 2018 Michael Still and Aptira
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for nvmetcli related routines.
"""

from oslo_concurrency import processutils

import cinder.privsep


@cinder.privsep.sys_admin_pctxt.entrypoint
def restore(path):
    cmd = [
        'nvmetcli',
        'restore',
        path]
    return processutils.execute(*cmd)


@cinder.privsep.sys_admin_pctxt.entrypoint
def save(path):
    cmd = [
        'nvmetcli',
        'save',
        path]
    return processutils.execute(*cmd)
